"""StarGAN training utilities."""

from __future__ import annotations

from dataclasses import dataclass

import torch
import torch.nn.functional as F

from src.models.stargan import StarGANDiscriminator, StarGANGenerator


@dataclass
class StarGANTrainingConfig:
    """Configuration for StarGAN adversarial training."""

    image_size: int = 128
    c_dim: int = 5
    g_conv_dim: int = 64
    d_conv_dim: int = 64
    g_repeat_num: int = 6
    d_repeat_num: int = 6
    lambda_cls: float = 1.0
    lambda_rec: float = 10.0
    lambda_gp: float = 10.0
    n_critic: int = 5
    g_lr: float = 1e-4
    d_lr: float = 1e-4
    beta1: float = 0.5
    beta2: float = 0.999
    device: str = "cpu"


class StarGANTrainer:
    """Training harness with WGAN-GP + domain classification + reconstruction."""

    def __init__(self, config: StarGANTrainingConfig) -> None:
        self.config = config
        self.device = torch.device(config.device)

        self.G = StarGANGenerator(
            conv_dim=config.g_conv_dim,
            c_dim=config.c_dim,
            repeat_num=config.g_repeat_num,
        ).to(self.device)
        self.D = StarGANDiscriminator(
            image_size=config.image_size,
            conv_dim=config.d_conv_dim,
            c_dim=config.c_dim,
            repeat_num=config.d_repeat_num,
        ).to(self.device)

        self.g_optimizer = torch.optim.Adam(
            self.G.parameters(),
            lr=config.g_lr,
            betas=(config.beta1, config.beta2),
        )
        self.d_optimizer = torch.optim.Adam(
            self.D.parameters(),
            lr=config.d_lr,
            betas=(config.beta1, config.beta2),
        )

    @staticmethod
    def _classification_loss(logit: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        # BCE-with-logits over multi-hot domain attributes.
        return F.binary_cross_entropy_with_logits(logit, target, reduction="sum") / logit.size(0)

    def _gradient_penalty(self, y: torch.Tensor, x: torch.Tensor) -> torch.Tensor:
        weight = torch.ones_like(y, device=self.device)
        dydx = torch.autograd.grad(
            outputs=y,
            inputs=x,
            grad_outputs=weight,
            retain_graph=True,
            create_graph=True,
            only_inputs=True,
        )[0]
        dydx = dydx.view(dydx.size(0), -1)
        return torch.mean((torch.norm(dydx, dim=1) - 1) ** 2)

    def train_discriminator(
        self,
        real_x: torch.Tensor,
        real_labels: torch.Tensor,
        target_labels: torch.Tensor,
    ) -> dict[str, float]:
        """One discriminator update."""
        real_x = real_x.to(self.device)
        real_labels = real_labels.to(self.device)
        target_labels = target_labels.to(self.device)

        out_src, out_cls = self.D(real_x)
        d_loss_real = -torch.mean(out_src)
        d_loss_cls = self._classification_loss(out_cls, real_labels)

        with torch.no_grad():
            fake_x = self.G(real_x, target_labels)
        out_src_fake, _ = self.D(fake_x.detach())
        d_loss_fake = torch.mean(out_src_fake)

        alpha = torch.rand(real_x.size(0), 1, 1, 1, device=self.device)
        x_hat = (alpha * real_x.data + (1 - alpha) * fake_x.data).requires_grad_(True)
        out_src_hat, _ = self.D(x_hat)
        d_loss_gp = self._gradient_penalty(out_src_hat, x_hat)

        d_loss = (
            d_loss_real
            + d_loss_fake
            + self.config.lambda_cls * d_loss_cls
            + self.config.lambda_gp * d_loss_gp
        )

        self.d_optimizer.zero_grad()
        d_loss.backward()
        self.d_optimizer.step()

        return {
            "d_loss": float(d_loss.item()),
            "d_real": float(d_loss_real.item()),
            "d_fake": float(d_loss_fake.item()),
            "d_cls": float(d_loss_cls.item()),
            "d_gp": float(d_loss_gp.item()),
        }

    def train_generator(
        self,
        real_x: torch.Tensor,
        real_labels: torch.Tensor,
        target_labels: torch.Tensor,
    ) -> dict[str, float]:
        """One generator update."""
        real_x = real_x.to(self.device)
        real_labels = real_labels.to(self.device)
        target_labels = target_labels.to(self.device)

        fake_x = self.G(real_x, target_labels)
        out_src, out_cls = self.D(fake_x)
        g_loss_fake = -torch.mean(out_src)
        g_loss_cls = self._classification_loss(out_cls, target_labels)

        rec_x = self.G(fake_x, real_labels)
        g_loss_rec = torch.mean(torch.abs(real_x - rec_x))

        g_loss = g_loss_fake + self.config.lambda_rec * g_loss_rec + self.config.lambda_cls * g_loss_cls

        self.g_optimizer.zero_grad()
        g_loss.backward()
        self.g_optimizer.step()

        return {
            "g_loss": float(g_loss.item()),
            "g_fake": float(g_loss_fake.item()),
            "g_cls": float(g_loss_cls.item()),
            "g_rec": float(g_loss_rec.item()),
        }

    def train_step(
        self,
        real_x: torch.Tensor,
        real_labels: torch.Tensor,
        target_labels: torch.Tensor,
        step: int,
    ) -> dict[str, float]:
        """Run one full StarGAN optimization step."""
        logs = self.train_discriminator(real_x, real_labels, target_labels)
        if step % self.config.n_critic == 0:
            logs.update(self.train_generator(real_x, real_labels, target_labels))
        return logs



# Copyright (c) 2026 EarthBridge Team.
# Credits: Built on open-source libraries and papers acknowledged in README.md citations.

"""Tests for DiT backbone (SiTBackbone), LBM scheduler, LBM pipeline,
and diffusers UNet wrappers (BDBM, DBIM, CDTSDE, LBM).

All components are imported from ``src/`` (no duplicate code in examples/).
"""

import pytest
import torch
import numpy as np


# ---------------------------------------------------------------------------
# SiTBackbone tests
# ---------------------------------------------------------------------------


class TestSiTBackbone:
    def test_output_shape_concat(self):
        from src.models.dit.sit import SiTBackbone

        model = SiTBackbone(
            image_size=16, patch_size=2, in_channels=3,
            hidden_size=64, depth=2, num_heads=4,
            condition_mode="concat",
        )
        x = torch.randn(2, 3, 16, 16)
        t = torch.tensor([0.5, 0.8])
        xT = torch.randn(2, 3, 16, 16)
        out = model(x, t, xT=xT)
        assert out.shape == (2, 3, 16, 16)

    def test_output_shape_unconditional(self):
        from src.models.dit.sit import SiTBackbone

        model = SiTBackbone(
            image_size=16, patch_size=2, in_channels=3,
            hidden_size=64, depth=2, num_heads=4,
            condition_mode=None,
        )
        x = torch.randn(1, 3, 16, 16)
        t = torch.tensor([1.0])
        out = model(x, t)
        assert out.shape == (1, 3, 16, 16)

    def test_sit_configs(self):
        from src.models.dit.sit import SIT_CONFIGS

        assert "S" in SIT_CONFIGS
        assert "B" in SIT_CONFIGS
        assert "L" in SIT_CONFIGS
        assert "XL" in SIT_CONFIGS
        depth, hidden, heads = SIT_CONFIGS["S"]
        assert depth == 12
        assert hidden == 384
        assert heads == 6

    def test_import_from_models(self):
        from src.models import SiTBackbone, SIT_CONFIGS
        assert SiTBackbone is not None
        assert SIT_CONFIGS is not None

    def test_import_from_top_level(self):
        import src
        assert hasattr(src, "SiTBackbone")
        assert hasattr(src, "SIT_CONFIGS")

    def test_different_channels(self):
        from src.models.dit.sit import SiTBackbone

        model = SiTBackbone(
            image_size=16, patch_size=2, in_channels=1,
            hidden_size=64, depth=2, num_heads=4,
            condition_mode="concat",
        )
        x = torch.randn(1, 1, 16, 16)
        t = torch.tensor([0.5])
        xT = torch.randn(1, 1, 16, 16)
        out = model(x, t, xT=xT)
        assert out.shape == (1, 1, 16, 16)


# ---------------------------------------------------------------------------
# LBM Scheduler tests
# ---------------------------------------------------------------------------


class TestLBMScheduler:
    def test_instantiation(self):
        from src.schedulers.lbm import LBMScheduler
        s = LBMScheduler()
        assert s.num_train_timesteps == 1000
        assert s.bridge_noise_sigma == 0.001

    def test_set_timesteps(self):
        from src.schedulers.lbm import LBMScheduler
        s = LBMScheduler()
        s.set_timesteps(5)
        assert s.num_inference_steps == 5
        assert len(s.timesteps) == 5
        assert len(s.sigmas) == 5

    def test_get_sigmas(self):
        from src.schedulers.lbm import LBMScheduler
        s = LBMScheduler(num_train_timesteps=100)
        t = torch.tensor([0, 50])
        sigmas = s.get_sigmas(t, n_dim=4, device="cpu")
        assert sigmas.shape == (2, 1, 1, 1)

    def test_add_noise(self):
        from src.schedulers.lbm import LBMScheduler
        s = LBMScheduler(num_train_timesteps=100)
        x_target = torch.randn(2, 3, 8, 8)
        x_source = torch.randn(2, 3, 8, 8)
        t = torch.tensor([10, 50])
        noisy = s.add_noise(x_target, x_source, t)
        assert noisy.shape == x_target.shape

    def test_step(self):
        from src.schedulers.lbm import LBMScheduler
        s = LBMScheduler(num_train_timesteps=100)
        s.set_timesteps(5)
        sample = torch.randn(1, 3, 8, 8)
        model_output = torch.randn(1, 3, 8, 8)
        t = s.timesteps[0]
        result = s.step(model_output, t, sample)
        assert result.prev_sample.shape == sample.shape
        assert result.pred_original_sample.shape == sample.shape

    def test_sample_timesteps_uniform(self):
        from src.schedulers.lbm import LBMScheduler
        s = LBMScheduler(num_train_timesteps=100)
        t = s.sample_timesteps(10)
        assert t.shape == (10,)
        assert t.dtype == torch.long

    def test_import_from_top_level(self):
        import src
        assert hasattr(src, "LBMScheduler")
        assert hasattr(src, "LBMSchedulerOutput")


# ---------------------------------------------------------------------------
# LBM Pipeline tests (src/)
# ---------------------------------------------------------------------------


class TestLBMPipelineImports:
    def test_import_from_pipelines(self):
        from src.pipelines import LBMPipeline, LBMPipelineOutput
        assert LBMPipeline is not None
        assert LBMPipelineOutput is not None

    def test_import_from_top_level(self):
        import src
        assert hasattr(src, "LBMPipeline")
        assert hasattr(src, "LBMPipelineOutput")


# ---------------------------------------------------------------------------
# UNet wrapper + pipeline tests: BDBM
# ---------------------------------------------------------------------------


class TestBDBMPipeline:
    @pytest.fixture
    def pipeline(self):
        from src.models.unet.diffusers_wrappers import BDBMUNet
        from src.schedulers.bdbm import BDBMScheduler
        from src.pipelines.bdbm import BDBMPipeline

        unet = BDBMUNet(
            image_size=32, in_channels=3, model_channels=32,
            num_res_blocks=1, attention_resolutions=(), condition_mode="concat",
        )
        scheduler = BDBMScheduler(num_timesteps=100, sample_step=10, objective="noise")
        return BDBMPipeline(unet=unet, scheduler=scheduler)

    def test_unet_forward(self):
        from src.models.unet.diffusers_wrappers import BDBMUNet

        unet = BDBMUNet(
            image_size=32, in_channels=3, model_channels=32,
            num_res_blocks=1, attention_resolutions=(), condition_mode="concat",
        )
        x = torch.randn(2, 3, 32, 32)
        t = torch.tensor([50, 80])
        ctx = torch.randn(2, 3, 32, 32)
        out = unet(x, t, context=ctx)
        assert out.shape == (2, 3, 32, 32)

    def test_pipeline_b2a(self, pipeline):
        source = torch.randn(1, 3, 32, 32)
        result = pipeline(source, direction="b2a", output_type="pt")
        assert isinstance(result.images, torch.Tensor)

    def test_pipeline_a2b(self, pipeline):
        source = torch.randn(1, 3, 32, 32)
        result = pipeline(source, direction="a2b", output_type="pt")
        assert isinstance(result.images, torch.Tensor)

    def test_pipeline_pil_output(self, pipeline):
        source = torch.randn(1, 3, 32, 32)
        result = pipeline(source, direction="b2a", output_type="pil")
        assert isinstance(result.images, list)
        assert len(result.images) == 1

    def test_pipeline_invalid_direction(self, pipeline):
        source = torch.randn(1, 3, 32, 32)
        with pytest.raises(ValueError, match="Unknown direction"):
            pipeline(source, direction="invalid", output_type="pt")


# ---------------------------------------------------------------------------
# UNet wrapper + pipeline tests: DBIM
# ---------------------------------------------------------------------------


class TestDBIMPipeline:
    @pytest.fixture
    def pipeline(self):
        from src.models.unet.diffusers_wrappers import DBIMUNet
        from src.schedulers.dbim import DBIMScheduler
        from src.pipelines.dbim import DBIMPipeline

        unet = DBIMUNet(
            image_size=32, in_channels=3, model_channels=32,
            num_res_blocks=1, attention_resolutions=(), condition_mode="concat",
        )
        scheduler = DBIMScheduler()
        return DBIMPipeline(unet=unet, scheduler=scheduler)

    def test_unet_forward(self):
        from src.models.unet.diffusers_wrappers import DBIMUNet

        unet = DBIMUNet(
            image_size=32, in_channels=3, model_channels=32,
            num_res_blocks=1, attention_resolutions=(), condition_mode="concat",
        )
        x = torch.randn(2, 3, 32, 32)
        t = torch.tensor([100.0, 200.0])
        xT = torch.randn(2, 3, 32, 32)
        out = unet(x, t, xT=xT)
        assert out.shape == (2, 3, 32, 32)

    def test_pipeline_pt_output(self, pipeline):
        source = torch.randn(1, 3, 32, 32)
        result = pipeline(source, num_inference_steps=3, output_type="pt")
        assert isinstance(result.images, torch.Tensor)
        assert result.nfe > 0

    def test_pipeline_pil_output(self, pipeline):
        source = torch.randn(1, 3, 32, 32)
        result = pipeline(source, num_inference_steps=3, output_type="pil")
        assert isinstance(result.images, list)
        assert len(result.images) == 1


# ---------------------------------------------------------------------------
# UNet wrapper + pipeline tests: CDTSDE
# ---------------------------------------------------------------------------


class TestCDTSDEPipeline:
    @pytest.fixture
    def pipeline(self):
        from src.models.unet.diffusers_wrappers import CDTSDEUNet
        from src.schedulers.cdtsde import CDTSDEScheduler
        from src.pipelines.cdtsde import CDTSDEPipeline

        unet = CDTSDEUNet(
            image_size=32, in_channels=3, model_channels=32,
            num_res_blocks=1, attention_resolutions=(), condition_mode="concat",
        )
        scheduler = CDTSDEScheduler(num_train_timesteps=100)
        return CDTSDEPipeline(unet=unet, scheduler=scheduler)

    def test_unet_forward(self):
        from src.models.unet.diffusers_wrappers import CDTSDEUNet

        unet = CDTSDEUNet(
            image_size=32, in_channels=3, model_channels=32,
            num_res_blocks=1, attention_resolutions=(), condition_mode="concat",
        )
        x = torch.randn(2, 3, 32, 32)
        t = torch.tensor([10, 50])
        xT = torch.randn(2, 3, 32, 32)
        out = unet(x, t, xT=xT)
        assert out.shape == (2, 3, 32, 32)

    def test_pipeline_pt_output(self, pipeline):
        source = torch.randn(1, 3, 32, 32)
        result = pipeline(source, num_inference_steps=5, output_type="pt")
        assert isinstance(result.images, torch.Tensor)
        assert result.nfe > 0

    def test_pipeline_pil_output(self, pipeline):
        source = torch.randn(1, 3, 32, 32)
        result = pipeline(source, num_inference_steps=5, output_type="pil")
        assert isinstance(result.images, list)
        assert len(result.images) == 1


# ---------------------------------------------------------------------------
# UNet wrapper + pipeline tests: LBM
# ---------------------------------------------------------------------------


class TestLBMPipeline:
    @pytest.fixture
    def pipeline(self):
        from src.models.unet.diffusers_wrappers import LBMUNet
        from src.schedulers.lbm import LBMScheduler
        from src.pipelines.lbm import LBMPipeline

        unet = LBMUNet(
            image_size=32, in_channels=3, model_channels=32,
            num_res_blocks=1, attention_resolutions=(), condition_mode="concat",
        )
        scheduler = LBMScheduler()
        return LBMPipeline(unet=unet, scheduler=scheduler)

    def test_unet_forward(self):
        from src.models.unet.diffusers_wrappers import LBMUNet

        unet = LBMUNet(
            image_size=32, in_channels=3, model_channels=32,
            num_res_blocks=1, attention_resolutions=(), condition_mode="concat",
        )
        x = torch.randn(2, 3, 32, 32)
        t = torch.tensor([0, 100])
        cond = torch.randn(2, 3, 32, 32)
        out = unet(x, t, cond=cond)
        assert out.shape == (2, 3, 32, 32)

    def test_scheduler_set_timesteps(self):
        from src.schedulers.lbm import LBMScheduler

        scheduler = LBMScheduler()
        scheduler.set_timesteps(5)
        assert len(scheduler.timesteps) == 5

    def test_pipeline_pt_output(self, pipeline):
        source = torch.randn(1, 3, 32, 32)
        result = pipeline(source, num_inference_steps=2, output_type="pt")
        assert isinstance(result.images, torch.Tensor)
        assert result.nfe > 0

    def test_pipeline_pil_output(self, pipeline):
        source = torch.randn(1, 3, 32, 32)
        result = pipeline(source, num_inference_steps=2, output_type="pil")
        assert isinstance(result.images, list)
        assert len(result.images) == 1

    def test_pipeline_np_output(self, pipeline):
        source = torch.randn(1, 3, 32, 32)
        result = pipeline(source, num_inference_steps=2, output_type="np")
        assert isinstance(result.images, np.ndarray)



"""Local Diffusion (Hallucination-aware diffusion) model components."""

from src.models.local_diffusion.local_diffusion_model import (
    LocalDiffusionUNet,
    ConditionEncoder,
    create_unet,
)

__all__ = [
    "LocalDiffusionUNet",
    "ConditionEncoder",
    "create_unet",
]



"""Image translation model architectures."""

from src.models.discriminators import PatchGANDiscriminator
from src.models.generators import ResNetGenerator, UNetGenerator
from src.models.unet import I2SBUNet
from src.models.unet.diffusers_wrappers import (
    BBDMUNet,
    BDBMUNet,
    BiBBDMUNet,
    CDTSDEUNet,
    DBIMUNet,
    DDBMUNet,
    DDIBUNet,
    I2SBDiffusersUNet,
    LBMUNet,
)
from src.models.dit import SiTBackbone, SIT_CONFIGS
from src.models.stegogan import (
    ResnetMaskV1Generator,
    ResnetMaskV3Generator,
    NetMatchability,
)
from src.models.cut import (
    CUTGenerator,
    create_generator,
    create_discriminator,
    create_patch_sample_mlp,
)
from src.models.unsb import (
    UNSBGenerator,
    UNSBDiscriminator,
    UNSBEnergyNet,
)
from src.models.local_diffusion import (
    LocalDiffusionUNet,
    ConditionEncoder,
)
from src.models.pix2pixhd import Pix2PixHDGenerator, Pix2PixHDGlobalGenerator
from src.models.stargan import StarGANGenerator, StarGANDiscriminator, StarGANResidualBlock

__all__ = [
    "UNetGenerator",
    "ResNetGenerator",
    "PatchGANDiscriminator",
    "I2SBUNet",
    "BBDMUNet",
    "BDBMUNet",
    "BiBBDMUNet",
    "CDTSDEUNet",
    "DBIMUNet",
    "DDBMUNet",
    "DDIBUNet",
    "I2SBDiffusersUNet",
    "LBMUNet",
    "SiTBackbone",
    "SIT_CONFIGS",
    "ResnetMaskV1Generator",
    "ResnetMaskV3Generator",
    "NetMatchability",
    "CUTGenerator",
    "create_generator",
    "create_discriminator",
    "create_patch_sample_mlp",
    "UNSBGenerator",
    "UNSBDiscriminator",
    "UNSBEnergyNet",
    "LocalDiffusionUNet",
    "ConditionEncoder",
    "Pix2PixHDGenerator",
    "Pix2PixHDGlobalGenerator",
    "StarGANGenerator",
    "StarGANDiscriminator",
    "StarGANResidualBlock",
]



# Copyright (c) 2026 EarthBridge Team.
# Credits: Built on open-source libraries and papers acknowledged in README.md citations.

"""DBIM inference pipeline for image-to-image translation.

Implements the Diffusion Bridge Implicit Models (DBIM) sampling loop,
which uses the same bridge model family as DDBM but introduces faster
samplers.

Reference
---------
Zheng, K., et al. "Diffusion Bridge Implicit Models." 2024.
"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, List, Optional, Tuple, Union

import numpy as np
import torch
from PIL import Image
from tqdm.auto import tqdm

from diffusers import DiffusionPipeline
from diffusers.utils import BaseOutput
from diffusers.utils.torch_utils import randn_tensor

from src.models.unet.diffusers_wrappers import DBIMUNet
from src.schedulers.dbim import DBIMScheduler


@dataclass
class DBIMPipelineOutput(BaseOutput):
    """Output of the DBIM pipeline.

    Attributes
    ----------
    images : list[PIL.Image.Image] | np.ndarray | torch.Tensor
        Translated images.
    nfe : int
        Number of function evaluations used.
    sampler : str
        Sampler used (``"dbim"`` or ``"heun"``).
    """

    images: Any
    nfe: int = 0
    sampler: str = "dbim"


class DBIMPipeline(DiffusionPipeline):
    """Image-to-image translation pipeline using DBIM.

    Parameters
    ----------
    unet : torch.nn.Module
        Trained denoising model.
    scheduler : DBIMScheduler
        DBIM noise scheduler.
    """

    model_cpu_offload_seq = "unet"

    def __init__(self, unet: torch.nn.Module, scheduler: DBIMScheduler) -> None:
        super().__init__()
        self.register_modules(unet=unet, scheduler=scheduler)

    @classmethod
    def from_pretrained(
        cls,
        pretrained_model_name_or_path: str | Path,
        *,
        subfolder: str = "unet",
        scheduler_subfolder: str = "scheduler",
        device: str | torch.device = "cpu",
        torch_dtype: torch.dtype | None = None,
        **kwargs,
    ) -> "DBIMPipeline":
        """Load DBIM pipeline from local checkpoint directories."""
        unet = DBIMUNet.from_pretrained(pretrained_model_name_or_path, subfolder=subfolder)
        try:
            scheduler = DBIMScheduler.from_pretrained(
                pretrained_model_name_or_path, subfolder=scheduler_subfolder
            )
        except Exception:
            scheduler = DBIMScheduler()
        unet = unet.eval().to(device=device)
        if torch_dtype is not None:
            unet = unet.to(dtype=torch_dtype)
        return cls(unet=unet, scheduler=scheduler)

    @property
    def device(self) -> torch.device:
        unet = self.unet.module if hasattr(self.unet, "module") else self.unet
        return next(unet.parameters()).device

    @property
    def dtype(self) -> torch.dtype:
        unet = self.unet.module if hasattr(self.unet, "module") else self.unet
        return next(unet.parameters()).dtype

    @staticmethod
    def _append_dims(x: torch.Tensor, target_dims: int) -> torch.Tensor:
        dims = target_dims - x.ndim
        return x[(...,) + (None,) * dims]

    def _bridge_scalings(
        self, t: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """DBIM bridge preconditioning ``(c_skip, c_out, c_in, c_noise)``."""
        a_t, b_t, c_t = self.scheduler.get_abc(t)
        sigma_data = self.scheduler.config.sigma_data

        A = a_t**2 * sigma_data**2 + b_t**2 * sigma_data**2 + c_t**2
        c_in = torch.rsqrt(torch.clamp(A, min=1e-20))
        c_skip = (b_t * sigma_data**2) / torch.clamp(A, min=1e-20)
        c_out = (
            torch.sqrt(torch.clamp(
                a_t**2 * sigma_data**4 + sigma_data**2 * c_t**2, min=1e-20,
            )) * c_in
        )
        c_noise = 1000.0 * 0.25 * torch.log(torch.clamp(t, min=1e-44))
        return c_skip, c_out, c_in, c_noise

    def denoise(
        self,
        x_t: torch.Tensor,
        t: torch.Tensor,
        x_T: torch.Tensor,
        clip_denoised: bool = True,
    ) -> torch.Tensor:
        """Predict denoised bridge state from noisy sample at time ``t``."""
        c_skip, c_out, c_in, c_noise = self._bridge_scalings(t)
        c_skip = self._append_dims(c_skip, x_t.ndim).to(dtype=x_t.dtype, device=x_t.device)
        c_out = self._append_dims(c_out, x_t.ndim).to(dtype=x_t.dtype, device=x_t.device)
        c_in = self._append_dims(c_in, x_t.ndim).to(dtype=x_t.dtype, device=x_t.device)

        model_output = self.unet(
            (c_in * x_t).to(dtype=self.dtype, device=self.device),
            c_noise.to(dtype=self.dtype, device=self.device),
            xT=x_T.to(dtype=self.dtype, device=self.device),
        )
        model_output = model_output.to(dtype=x_t.dtype, device=x_t.device)
        denoised = c_out * model_output + c_skip * x_t
        if clip_denoised:
            denoised = denoised.clamp(-1, 1)
        return denoised

    @staticmethod
    def prepare_inputs(image, device, dtype):
        """Convert inputs to ``[-1, 1]`` tensors."""
        if isinstance(image, Image.Image):
            image = [image]
        if isinstance(image, list) and isinstance(image[0], Image.Image):
            images = []
            for img in image:
                img = img.convert("RGB")
                arr = np.array(img).astype(np.float32) / 255.0
                images.append(torch.from_numpy(arr).permute(2, 0, 1))
            image = torch.stack(images)
        if isinstance(image, np.ndarray):
            image = torch.from_numpy(image)
        if image.max() > 1.0:
            image = image / 255.0
        if image.min() >= 0:
            image = image * 2 - 1
        return image.to(device=device, dtype=dtype)

    @torch.no_grad()
    def __call__(
        self,
        source_image: Union[torch.Tensor, Image.Image, List[Image.Image]],
        num_inference_steps: int = 20,
        output_type: str = "pil",
        return_dict: bool = True,
        generator: Optional[torch.Generator] = None,
        callback: Optional[Callable] = None,
        callback_steps: int = 1,
    ):
        """Run DBIM bridge diffusion sampling.

        Parameters
        ----------
        source_image : Tensor | PIL.Image | list[PIL.Image]
            Source image(s).
        num_inference_steps : int
            Number of diffusion steps.
        output_type : str
            ``"pil"`` | ``"np"`` | ``"pt"``.
        """
        device = self.device
        dtype = self.dtype
        x_T = self.prepare_inputs(source_image, device, dtype)
        batch_size = x_T.shape[0]

        self.scheduler.set_timesteps(num_inference_steps, device=device)
        sigmas = self.scheduler.sigmas
        s_in = x_T.new_ones([batch_size])
        nfe = 0

        # Initialise from bridge marginal at t=sigma_max
        a_0, b_0, c_0 = [
            self._append_dims(v, x_T.ndim) for v in self.scheduler.get_abc(sigmas[0] * s_in)
        ]
        noise = randn_tensor(x_T.shape, generator=generator, device=device, dtype=dtype)
        x = a_0 * x_T + c_0 * noise

        for i in tqdm(range(len(sigmas) - 1), desc="DBIM Sampling"):
            sigma = sigmas[i]
            denoised = self.denoise(x, sigma * s_in, x_T)
            nfe += 1

            result = self.scheduler.step(
                model_output=denoised,
                timestep=i,
                sample=x,
                x_T=x_T,
                generator=generator,
            )
            x = result.prev_sample

            if callback is not None and i % callback_steps == 0:
                callback(i, num_inference_steps, x)

        images = x.clamp(-1, 1)
        if output_type == "pil":
            images = self._convert_to_pil(images)
        elif output_type == "np":
            images = self._convert_to_numpy(images)

        if not return_dict:
            return (images, nfe)
        return DBIMPipelineOutput(images=images, nfe=nfe, sampler="dbim")

    @staticmethod
    def _convert_to_pil(images: torch.Tensor) -> List[Image.Image]:
        images = (images + 1) / 2
        images = images.clamp(0, 1).cpu().permute(0, 2, 3, 1).numpy()
        images = (images * 255).round().astype(np.uint8)
        return [Image.fromarray(img) for img in images]

    @staticmethod
    def _convert_to_numpy(images: torch.Tensor) -> np.ndarray:
        images = (images + 1) / 2
        return images.clamp(0, 1).cpu().permute(0, 2, 3, 1).numpy()



# Copyright (c) 2026 EarthBridge Team.
# Credits: Adapted from Parallel-GAN (Wang et al., TGRS 2022).

"""Training configuration and harness for Parallel-GAN."""

from __future__ import annotations

import functools
from dataclasses import dataclass

import torch
import torch.nn as nn

from examples.community.parallel_gan.model import (
    ParaGAN,
    Resrecon,
    VGGLoss,
    _GANLoss,
    _NLayerDiscriminator,
    _init_weights,
)


# ---------------------------------------------------------------------------
# Configuration
# ---------------------------------------------------------------------------


@dataclass
class ParallelGANConfig:
    """Configuration for :class:`ParallelGANTrainer`.

    Parameters
    ----------
    input_nc : int
        SAR input channels (3 or 4).
    output_nc : int
        Optical output channels.
    ngf : int
        Generator feature width.
    ndf : int
        Discriminator feature width.
    n_layers_D : int
        Discriminator layers.
    norm : str
        ``"batch"`` or ``"instance"``.
    gan_mode : str
        ``"vanilla"`` or ``"lsgan"``.
    lambda_L1 : float
        Weight for pixel-wise L1 loss.
    lambda_vgg : float
        Weight for VGG perceptual loss (Stage 1 only).
    lambda_feat : float
        Weight for hierarchical feature loss (Stage 2 only).
    feature_weights : tuple[float, ...]
        Per-level weights for the feature loss (5 levels).
    n_blocks : int
        Number of residual blocks in the translation generator.
    lr : float
        Learning rate.
    beta1 : float
        Adam beta1.
    device : str
        ``"cpu"`` or ``"cuda"``.
    """

    input_nc: int = 3
    output_nc: int = 3
    ngf: int = 64
    ndf: int = 64
    n_layers_D: int = 3
    norm: str = "batch"
    gan_mode: str = "vanilla"
    lambda_L1: float = 100.0
    lambda_vgg: float = 10.0
    lambda_feat: float = 10.0
    feature_weights: tuple[float, ...] = (1 / 32, 1 / 16, 1 / 8, 1 / 4, 1 / 2)
    n_blocks: int = 6
    lr: float = 2e-4
    beta1: float = 0.5
    device: str = "cpu"


# ---------------------------------------------------------------------------
# Trainer
# ---------------------------------------------------------------------------


class ParallelGANTrainer:
    """Two-stage training harness for Parallel-GAN.

    Instantiate with a :class:`ParallelGANConfig`, then call
    :meth:`train_step_recon` for Stage 1 (reconstruction) or
    :meth:`train_step_trans` for Stage 2 (translation).

    Parameters
    ----------
    config : ParallelGANConfig
        Training configuration.
    recon_net : Resrecon, optional
        Pre-trained reconstruction network for Stage 2 feature supervision.
        If ``None``, only Stage 1 is available.
    """

    def __init__(
        self,
        config: ParallelGANConfig,
        recon_net: Resrecon | None = None,
    ) -> None:
        self.config = config
        self.device = torch.device(config.device)

        norm_layer: type[nn.Module]
        if config.norm == "batch":
            norm_layer = nn.BatchNorm2d
        else:
            norm_layer = functools.partial(nn.InstanceNorm2d, affine=False, track_running_stats=False)  # type: ignore[assignment]

        # Generator (translation)
        self.netG = ParaGAN(
            input_nc=config.input_nc,
            output_nc=config.output_nc,
            norm_layer=norm_layer,
            n_blocks=config.n_blocks,
        ).to(self.device)
        _init_weights(self.netG)

        # Discriminator
        self.netD = _NLayerDiscriminator(
            input_nc=config.input_nc + config.output_nc,
            ndf=config.ndf,
            n_layers=config.n_layers_D,
            norm_layer=norm_layer,
        ).to(self.device)
        _init_weights(self.netD)

        # Reconstruction network (frozen, for Stage 2)
        self.recon_net = recon_net
        if self.recon_net is not None:
            self.recon_net = self.recon_net.to(self.device)
            self.recon_net.eval()
            for p in self.recon_net.parameters():
                p.requires_grad = False

        # Losses
        self.criterion_GAN = _GANLoss(config.gan_mode).to(self.device)
        self.criterion_L1 = nn.L1Loss()
        self.criterion_VGG = VGGLoss().to(self.device)

        # Optimisers
        self.optimizer_G = torch.optim.Adam(
            self.netG.parameters(), lr=config.lr, betas=(config.beta1, 0.999)
        )
        self.optimizer_D = torch.optim.Adam(
            self.netD.parameters(), lr=config.lr, betas=(config.beta1, 0.999)
        )

    # ----- Stage 1: reconstruction -----------------------------------------

    def train_step_recon(
        self,
        real_B: torch.Tensor,
    ) -> dict[str, float]:
        """One reconstruction training step (Stage 1).

        The reconstruction network learns to reconstruct optical images
        from optical images.

        Parameters
        ----------
        real_B : Tensor ``[B, C, H, W]``
            Optical images.
        """
        real_B = real_B.to(self.device)
        features = self.netG(real_B)
        fake_B = features[-1]

        # Discriminator
        self._set_requires_grad(self.netD, True)
        self.optimizer_D.zero_grad()
        loss_D = self._backward_D(real_B, fake_B, real_B)
        self.optimizer_D.step()

        # Generator
        self._set_requires_grad(self.netD, False)
        self.optimizer_G.zero_grad()
        fake_AB = torch.cat([real_B, fake_B], dim=1)
        loss_G_GAN = self.criterion_GAN(self.netD(fake_AB), target_is_real=True)
        loss_G_L1 = self.criterion_L1(fake_B, real_B) * self.config.lambda_L1
        loss_VGG = self.criterion_VGG(fake_B, real_B) * self.config.lambda_vgg
        loss_G = loss_G_GAN + loss_G_L1 + loss_VGG
        loss_G.backward()
        self.optimizer_G.step()

        return {
            "D": loss_D,
            "G_GAN": loss_G_GAN.item(),
            "G_L1": loss_G_L1.item(),
            "VGG": loss_VGG.item(),
        }

    # ----- Stage 2: translation --------------------------------------------

    def train_step_trans(
        self,
        real_A: torch.Tensor,
        real_B: torch.Tensor,
    ) -> dict[str, float]:
        """One translation training step (Stage 2).

        Parameters
        ----------
        real_A : Tensor ``[B, C_in, H, W]``
            SAR images.
        real_B : Tensor ``[B, C_out, H, W]``
            Optical images.
        """
        if self.recon_net is None:
            raise RuntimeError(
                "Stage 2 requires a pre-trained reconstruction network. "
                "Pass recon_net= to ParallelGANTrainer."
            )

        real_A = real_A.to(self.device)
        real_B = real_B.to(self.device)

        trans_features = self.netG(real_A)
        fake_B = trans_features[-1]

        with torch.no_grad():
            recon_features = self.recon_net(real_B)

        # Discriminator
        self._set_requires_grad(self.netD, True)
        self.optimizer_D.zero_grad()
        loss_D = self._backward_D(real_A, fake_B, real_B)
        self.optimizer_D.step()

        # Generator
        self._set_requires_grad(self.netD, False)
        self.optimizer_G.zero_grad()
        fake_AB = torch.cat([real_A, fake_B], dim=1)
        loss_G_GAN = self.criterion_GAN(self.netD(fake_AB), target_is_real=True)
        loss_G_L1 = self.criterion_L1(fake_B, real_B) * self.config.lambda_L1

        loss_feat = torch.tensor(0.0, device=self.device)
        weights = self.config.feature_weights
        for i in range(len(trans_features) - 1):
            loss_feat = loss_feat + self.criterion_L1(
                trans_features[i], recon_features[i]
            ) * weights[i] * self.config.lambda_feat

        loss_G = loss_G_GAN + loss_G_L1 + loss_feat
        loss_G.backward()
        self.optimizer_G.step()

        return {
            "D": loss_D,
            "G_GAN": loss_G_GAN.item(),
            "G_L1": loss_G_L1.item(),
            "feat": loss_feat.item(),
        }

    # ----- Convenience: single step (auto-selects stage) -------------------

    def train_step(
        self,
        real_A: torch.Tensor,
        real_B: torch.Tensor,
    ) -> dict[str, float]:
        """Run one training step.

        If a reconstruction network was provided, runs Stage 2 (translation).
        Otherwise, runs Stage 1 (reconstruction) using *real_B* only.
        """
        if self.recon_net is not None:
            return self.train_step_trans(real_A, real_B)
        return self.train_step_recon(real_B)

    # ----- Helpers ----------------------------------------------------------

    def _backward_D(
        self,
        real_A: torch.Tensor,
        fake_B: torch.Tensor,
        real_B: torch.Tensor,
    ) -> float:
        fake_AB = torch.cat([real_A, fake_B.detach()], dim=1)
        real_AB = torch.cat([real_A, real_B], dim=1)
        loss_fake = self.criterion_GAN(self.netD(fake_AB), target_is_real=False)
        loss_real = self.criterion_GAN(self.netD(real_AB), target_is_real=True)
        loss_D = (loss_fake + loss_real) * 0.5
        loss_D.backward()
        return loss_D.item()

    @staticmethod
    def _set_requires_grad(net: nn.Module, requires_grad: bool) -> None:
        for p in net.parameters():
            p.requires_grad = requires_grad



# Copyright (c) 2026 EarthBridge Team.
# Credits: Built on open-source libraries and papers acknowledged in README.md citations.

"""Local Diffusion model components built on native PyTorch modules.

This module provides the networks required by the Local Diffusion method
from *Kim et al. "Tackling Structural Hallucination in Image Translation
with Local Diffusion" (ECCV 2024, Oral)*:

* :class:`LocalDiffusionUNet` – Conditional denoising U-Net that fuses
  an independent condition encoder branch at the bottleneck.  The encoder
  (``ConditionEncoder``) extracts multi-scale features from the conditioning
  image (e.g. a degraded or corrupted input), and those features are
  concatenated with the diffusion latent at the bottleneck for fusion.
* :class:`ConditionEncoder` – Residual encoder for the conditioning image.
  Uses GroupNorm-based ``BasicBlock`` layers and max-pooling to produce a
  compact spatial feature map.

The architecture supports three beta schedules (linear, cosine, sigmoid)
and three prediction objectives (``pred_x0``, ``pred_noise``, ``pred_v``).
"""

from __future__ import annotations

import math
from functools import partial
from typing import Optional, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import init

from einops import rearrange
from einops.layers.torch import Rearrange


# ---------------------------------------------------------------------------
# Small helpers
# ---------------------------------------------------------------------------

def _exists(x):
    return x is not None


def _default(val, d):
    if _exists(val):
        return val
    return d() if callable(d) else d


def _divisible_by(numer: int, denom: int) -> bool:
    return (numer % denom) == 0


def _cast_tuple(t, length: int = 1):
    return t if isinstance(t, tuple) else (t,) * length


# ---------------------------------------------------------------------------
# Upsample / Downsample
# ---------------------------------------------------------------------------

def _upsample(dim: int, dim_out: Optional[int] = None) -> nn.Sequential:
    return nn.Sequential(
        nn.Upsample(scale_factor=2, mode="nearest"),
        nn.Conv2d(dim, _default(dim_out, dim), 3, padding=1),
    )


def _downsample(dim: int, dim_out: Optional[int] = None) -> nn.Sequential:
    return nn.Sequential(
        Rearrange("b c (h p1) (w p2) -> b (c p1 p2) h w", p1=2, p2=2),
        nn.Conv2d(dim * 4, _default(dim_out, dim), 1),
    )


# ---------------------------------------------------------------------------
# RMSNorm
# ---------------------------------------------------------------------------

class _RMSNorm(nn.Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.g = nn.Parameter(torch.ones(1, dim, 1, 1))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.normalize(x, dim=1) * self.g * (x.shape[1] ** 0.5)


# ---------------------------------------------------------------------------
# Sinusoidal positional embedding
# ---------------------------------------------------------------------------

class _SinusoidalPosEmb(nn.Module):
    def __init__(self, dim: int, theta: float = 10000.0) -> None:
        super().__init__()
        self.dim = dim
        self.theta = theta

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        half_dim = self.dim // 2
        emb = math.log(self.theta) / (half_dim - 1)
        emb = torch.exp(torch.arange(half_dim, device=x.device) * -emb)
        emb = x[:, None] * emb[None, :]
        return torch.cat((emb.sin(), emb.cos()), dim=-1)


# ---------------------------------------------------------------------------
# Building blocks (Block, ResnetBlock, LinearAttention, Attention)
# ---------------------------------------------------------------------------

class _Block(nn.Module):
    def __init__(self, dim: int, dim_out: int, groups: int = 8) -> None:
        super().__init__()
        self.proj = nn.Conv2d(dim, dim_out, 3, padding=1)
        self.norm = nn.GroupNorm(groups, dim_out)
        self.act = nn.SiLU()

    def forward(self, x: torch.Tensor, scale_shift: Optional[Tuple] = None) -> torch.Tensor:
        x = self.proj(x)
        x = self.norm(x)
        if _exists(scale_shift):
            scale, shift = scale_shift
            x = x * (scale + 1) + shift
        return self.act(x)


class _ResnetBlock(nn.Module):
    """ResNet block with optional time-embedding injection (FiLM)."""

    def __init__(self, dim: int, dim_out: int, *, time_emb_dim: Optional[int] = None, groups: int = 8) -> None:
        super().__init__()
        self.mlp = (
            nn.Sequential(nn.SiLU(), nn.Linear(time_emb_dim, dim_out * 2))
            if _exists(time_emb_dim)
            else None
        )
        self.block1 = _Block(dim, dim_out, groups=groups)
        self.block2 = _Block(dim_out, dim_out, groups=groups)
        self.res_conv = nn.Conv2d(dim, dim_out, 1) if dim != dim_out else nn.Identity()

    def forward(self, x: torch.Tensor, time_emb: Optional[torch.Tensor] = None) -> torch.Tensor:
        scale_shift = None
        if _exists(self.mlp) and _exists(time_emb):
            time_emb = self.mlp(time_emb)
            time_emb = rearrange(time_emb, "b c -> b c 1 1")
            scale_shift = time_emb.chunk(2, dim=1)
        h = self.block1(x, scale_shift=scale_shift)
        h = self.block2(h)
        return h + self.res_conv(x)


class _LinearAttention(nn.Module):
    """Fast O(N) linear attention."""

    def __init__(self, dim: int, heads: int = 4, dim_head: int = 32) -> None:
        super().__init__()
        self.scale = dim_head ** -0.5
        self.heads = heads
        hidden_dim = dim_head * heads
        self.norm = _RMSNorm(dim)
        self.to_qkv = nn.Conv2d(dim, hidden_dim * 3, 1, bias=False)
        self.to_out = nn.Sequential(nn.Conv2d(hidden_dim, dim, 1), _RMSNorm(dim))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        b, c, h, w = x.shape
        x = self.norm(x)
        qkv = self.to_qkv(x).chunk(3, dim=1)
        q, k, v = (rearrange(t, "b (h c) x y -> b h c (x y)", h=self.heads) for t in qkv)
        q = q.softmax(dim=-2) * self.scale
        k = k.softmax(dim=-1)
        context = torch.einsum("b h d n, b h e n -> b h d e", k, v)
        out = torch.einsum("b h d e, b h d n -> b h e n", context, q)
        out = rearrange(out, "b h c (x y) -> b (h c) x y", h=self.heads, x=h, y=w)
        return self.to_out(out)


class _Attention(nn.Module):
    """Standard scaled dot-product attention."""

    def __init__(self, dim: int, heads: int = 4, dim_head: int = 32) -> None:
        super().__init__()
        self.heads = heads
        hidden_dim = dim_head * heads
        self.norm = _RMSNorm(dim)
        self.to_qkv = nn.Conv2d(dim, hidden_dim * 3, 1, bias=False)
        self.to_out = nn.Conv2d(hidden_dim, dim, 1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        b, c, h, w = x.shape
        x = self.norm(x)
        qkv = self.to_qkv(x).chunk(3, dim=1)
        q, k, v = (rearrange(t, "b (h c) x y -> b h (x y) c", h=self.heads) for t in qkv)
        scale = q.shape[-1] ** -0.5
        sim = torch.einsum("b h i d, b h j d -> b h i j", q, k) * scale
        attn = sim.softmax(dim=-1)
        out = torch.einsum("b h i j, b h j d -> b h i d", attn, v)
        out = rearrange(out, "b h (x y) d -> b (h d) x y", x=h, y=w)
        return self.to_out(out)


# ---------------------------------------------------------------------------
# ConditionEncoder (ResUnet-style feature encoder)
# ---------------------------------------------------------------------------

class _CondBasicBlock(nn.Module):
    """GroupNorm-based residual block for the condition encoder."""

    def __init__(self, in_ch: int, mid_ch: int, out_ch: int, stride: int = 1, residual: bool = True) -> None:
        super().__init__()
        self.residual = residual
        self.in_ch = in_ch
        self.out_ch = out_ch

        def _gn(ch: int) -> nn.GroupNorm:
            """GroupNorm with adaptive group count (max 16, must divide ch)."""
            groups = min(16, ch)
            while ch % groups != 0:
                groups -= 1
            return nn.GroupNorm(groups, ch)

        self.convblock = nn.Sequential(
            nn.Conv2d(in_ch, mid_ch, kernel_size=3, stride=stride, padding=1),
            _gn(mid_ch),
            nn.ReLU(),
            nn.Conv2d(mid_ch, out_ch, kernel_size=3, padding=1),
            _gn(out_ch),
        )
        self.identity = None
        if residual and in_ch != out_ch:
            self.identity = nn.Sequential(
                nn.Conv2d(in_ch, out_ch, kernel_size=3, stride=stride, padding=1),
                _gn(out_ch),
            )
        self.relu = nn.ReLU()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x
        out = self.convblock(x)
        if self.residual:
            if self.identity is not None:
                identity = self.identity(x)
            out = out + identity
        return self.relu(out)


class ConditionEncoder(nn.Module):
    """Residual encoder for the conditioning image.

    Extracts multi-scale features and outputs a compact spatial feature
    map at the bottleneck resolution.  The architecture mirrors a ResUnet
    encoder with GroupNorm.

    The encoder downsamples using max-pooling.  The number of max-pool
    operations is ``len(filters) - 2`` (i.e. one fewer than the number
    of blocks).  Set *filters* to match the U-Net's depth so that the
    encoder output spatial size equals the U-Net bottleneck spatial size.

    Parameters
    ----------
    in_channels : int
        Number of input channels (e.g. 1 for grayscale, 3 for RGB).
    filters : tuple of int
        Channel widths at each encoder stage.  The number of downsampling
        operations equals ``len(filters) - 2``.
    """

    def __init__(
        self,
        in_channels: int = 1,
        filters: Tuple[int, ...] = (32, 32, 64, 128, 256),
    ) -> None:
        super().__init__()
        self.blocks = nn.ModuleList()
        ch_in = in_channels
        for ch_out in filters:
            self.blocks.append(_CondBasicBlock(ch_in, ch_out, ch_out))
            ch_in = ch_out
        self.maxpool = nn.MaxPool2d(2)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        for i, block in enumerate(self.blocks):
            x = block(x)
            # Pool between blocks (not after the last one)
            if i < len(self.blocks) - 1:
                x = self.maxpool(x)
        return x


# ---------------------------------------------------------------------------
# LocalDiffusionUNet – conditional denoising U-Net
# ---------------------------------------------------------------------------

class LocalDiffusionUNet(nn.Module):
    """Conditional denoising U-Net for Local Diffusion.

    The U-Net accepts a noisy image ``x_t`` and a timestep ``t``.  A separate
    :class:`ConditionEncoder` processes the conditioning image and its output
    is concatenated with the U-Net bottleneck features, followed by a fusion
    block.

    Parameters
    ----------
    dim : int
        Base channel dimension.
    channels : int
        Number of image channels (input and output).
    dim_mults : tuple of int
        Channel multipliers for each encoder/decoder level.
    resnet_block_groups : int
        Number of groups for GroupNorm in ResNet blocks.
    attn_dim_head : int
        Dimension per attention head.
    attn_heads : int
        Number of attention heads.
    full_attn : tuple of bool
        Per-level toggle for full (quadratic) vs. linear attention.
    cond_in_channels : int, optional
        Input channels for the conditioning encoder.  If ``None``, defaults
        to *channels*.
    cond_filters : tuple of int, optional
        Channel widths for the condition encoder.  If ``None``, automatically
        constructed to have the same number of pooling stages as the U-Net
        encoder, ensuring spatial size alignment at the bottleneck.
    init_type : str
        Weight initialisation method (``"normal"`` | ``"xavier"`` | ``"kaiming"``).
    init_gain : float
        Gain for weight initialisation.
    """

    def __init__(
        self,
        dim: int = 32,
        channels: int = 1,
        dim_mults: Tuple[int, ...] = (1, 2, 4, 8),
        resnet_block_groups: int = 8,
        attn_dim_head: int = 32,
        attn_heads: int = 4,
        full_attn: Tuple[bool, ...] = (False, False, False, True),
        cond_in_channels: Optional[int] = None,
        cond_filters: Optional[Tuple[int, ...]] = None,
        init_type: str = "normal",
        init_gain: float = 0.02,
    ) -> None:
        super().__init__()
        self.channels = channels
        cond_in_channels = cond_in_channels if cond_in_channels is not None else channels

        # Condition encoder
        # The U-Net spatially downsamples (len(dim_mults) - 1) times.
        # The cond encoder needs the same number of pooling ops, meaning
        # (n_pools + 1) blocks.  Its final output channels must equal
        # the U-Net bottleneck channels (dims[-1] = dim * dim_mults[-1]).
        if cond_filters is None:
            n_pools = len(dim_mults) - 1
            n_blocks = n_pools + 1
            # Build ascending channel widths up to mid_dim
            mid_dim_target = dim * dim_mults[-1]
            cond_filters_list = []
            for i in range(n_blocks):
                ch = max(dim, min(dim * (2 ** i), mid_dim_target))
                cond_filters_list.append(ch)
            cond_filters_list[-1] = mid_dim_target
            cond_filters = tuple(cond_filters_list)

        self.cond_model = ConditionEncoder(
            in_channels=cond_in_channels,
            filters=cond_filters,
        )

        # Initial projection
        init_dim = dim
        self.init_conv = nn.Conv2d(channels, init_dim, 7, padding=3)

        # Channel dimensions per level
        dims = [init_dim, *(dim * m for m in dim_mults)]
        in_out = list(zip(dims[:-1], dims[1:]))

        block_klass = partial(_ResnetBlock, groups=resnet_block_groups)

        # Time embedding
        time_dim = dim * 4
        sinu_pos_emb = _SinusoidalPosEmb(dim)
        self.time_mlp = nn.Sequential(
            sinu_pos_emb,
            nn.Linear(dim, time_dim),
            nn.GELU(),
            nn.Linear(time_dim, time_dim),
        )

        # Attention configs
        num_stages = len(dim_mults)
        full_attn = _cast_tuple(full_attn, num_stages)
        attn_heads = _cast_tuple(attn_heads, num_stages)
        attn_dim_head = _cast_tuple(attn_dim_head, num_stages)

        # Encoder
        self.downs = nn.ModuleList()
        for ind, ((dim_in, dim_out), layer_full_attn, layer_heads, layer_dim_head) in enumerate(
            zip(in_out, full_attn, attn_heads, attn_dim_head)
        ):
            is_last = ind >= (len(in_out) - 1)
            attn_klass = _Attention if layer_full_attn else _LinearAttention
            self.downs.append(
                nn.ModuleList([
                    block_klass(dim_in, dim_in, time_emb_dim=time_dim),
                    block_klass(dim_in, dim_in, time_emb_dim=time_dim),
                    attn_klass(dim_in, dim_head=layer_dim_head, heads=layer_heads),
                    _downsample(dim_in, dim_out) if not is_last else nn.Conv2d(dim_in, dim_out, 3, padding=1),
                ])
            )

        # Bottleneck
        mid_dim = dims[-1]
        self.mid_block1 = block_klass(mid_dim, mid_dim, time_emb_dim=time_dim)
        self.mid_attn = _Attention(mid_dim, heads=attn_heads[-1], dim_head=attn_dim_head[-1])
        self.mid_block2 = block_klass(mid_dim, mid_dim, time_emb_dim=time_dim)

        # Fusion block: concat UNet bottleneck + condition features → project back
        self.conv_fusion = block_klass(mid_dim * 2, mid_dim, time_emb_dim=time_dim)

        # Decoder
        self.ups = nn.ModuleList()
        for ind, ((dim_in, dim_out), layer_full_attn, layer_heads, layer_dim_head) in enumerate(
            zip(
                *map(reversed, (in_out, full_attn, attn_heads, attn_dim_head))
            )
        ):
            is_last = ind == (len(in_out) - 1)
            attn_klass = _Attention if layer_full_attn else _LinearAttention
            self.ups.append(
                nn.ModuleList([
                    block_klass(dim_out + dim_in, dim_out, time_emb_dim=time_dim),
                    block_klass(dim_out + dim_in, dim_out, time_emb_dim=time_dim),
                    attn_klass(dim_out, dim_head=layer_dim_head, heads=layer_heads),
                    _upsample(dim_out, dim_in) if not is_last else nn.Conv2d(dim_out, dim_in, 3, padding=1),
                ])
            )

        # Final projection
        self.final_res_block = block_klass(dim * 2, dim, time_emb_dim=time_dim)
        self.final_conv = nn.Conv2d(dim, channels, 1)

        _init_weights(self, init_type, init_gain)

    @property
    def downsample_factor(self) -> int:
        return 2 ** (len(self.downs) - 1)

    def forward(
        self,
        x: torch.Tensor,
        cond_img: torch.Tensor,
        time: torch.Tensor,
    ) -> torch.Tensor:
        """Forward pass.

        Parameters
        ----------
        x : Tensor (B, C, H, W)
            Noisy image at timestep ``t``.
        cond_img : Tensor (B, C_cond, H, W)
            Conditioning image (e.g. degraded / corrupted input).
        time : Tensor (B,)
            Integer or float timestep index.

        Returns
        -------
        Tensor (B, C, H, W)
            Model prediction (x_0, noise, or v depending on objective).
        """
        x = self.init_conv(x)
        r = x.clone()
        t = self.time_mlp(time)
        h = []

        # Encoder path
        for block1, block2, attn, downsample in self.downs:
            x = block1(x, t)
            h.append(x)
            x = block2(x, t)
            x = attn(x) + x
            h.append(x)
            x = downsample(x)

        # Bottleneck
        x = self.mid_block1(x, t)
        x = self.mid_attn(x) + x
        x = self.mid_block2(x, t)

        # Condition fusion
        cond_feat = self.cond_model(cond_img.to(torch.float))
        x = torch.cat((x, cond_feat), dim=1)
        x = self.conv_fusion(x, t)

        # Decoder path
        for block1, block2, attn, upsample in self.ups:
            x = torch.cat((x, h.pop()), dim=1)
            x = block1(x, t)
            x = torch.cat((x, h.pop()), dim=1)
            x = block2(x, t)
            x = attn(x) + x
            x = upsample(x)

        x = torch.cat((x, r), dim=1)
        x = self.final_res_block(x, t)
        return self.final_conv(x)


# ---------------------------------------------------------------------------
# Weight initialisation
# ---------------------------------------------------------------------------

def _init_weights(net: nn.Module, init_type: str = "normal", init_gain: float = 0.02) -> None:
    """Initialise network weights."""

    def _init_func(m: nn.Module) -> None:
        classname = m.__class__.__name__
        if hasattr(m, "weight") and (classname.find("Conv") != -1 or classname.find("Linear") != -1):
            if init_type == "normal":
                init.normal_(m.weight.data, 0.0, init_gain)
            elif init_type == "xavier":
                init.xavier_normal_(m.weight.data, gain=init_gain)
            elif init_type == "kaiming":
                init.kaiming_normal_(m.weight.data, a=0, mode="fan_in")
            elif init_type == "orthogonal":
                init.orthogonal_(m.weight.data, gain=init_gain)
            if hasattr(m, "bias") and m.bias is not None:
                init.constant_(m.bias.data, 0.0)
        elif classname.find("GroupNorm") != -1 or classname.find("BatchNorm") != -1:
            if hasattr(m, "weight") and m.weight is not None:
                init.normal_(m.weight.data, 1.0, init_gain)
            if hasattr(m, "bias") and m.bias is not None:
                init.constant_(m.bias.data, 0.0)

    net.apply(_init_func)


# ---------------------------------------------------------------------------
# Factory functions
# ---------------------------------------------------------------------------

def create_unet(
    dim: int = 32,
    channels: int = 1,
    dim_mults: Tuple[int, ...] = (1, 2, 4, 8),
    resnet_block_groups: int = 8,
    attn_dim_head: int = 32,
    attn_heads: int = 4,
    full_attn: Tuple[bool, ...] = (False, False, False, True),
    cond_in_channels: Optional[int] = None,
    cond_filters: Optional[Tuple[int, ...]] = None,
    init_type: str = "normal",
    init_gain: float = 0.02,
    **kwargs,
) -> LocalDiffusionUNet:
    """Factory for the Local Diffusion U-Net."""
    return LocalDiffusionUNet(
        dim=dim,
        channels=channels,
        dim_mults=dim_mults,
        resnet_block_groups=resnet_block_groups,
        attn_dim_head=attn_dim_head,
        attn_heads=attn_heads,
        full_attn=full_attn,
        cond_in_channels=cond_in_channels,
        cond_filters=cond_filters,
        init_type=init_type,
        init_gain=init_gain,
    )



"""StarGAN model architectures.

Adapted from the official implementation:
https://github.com/yunjey/stargan
"""

from src.models.stargan.blocks import StarGANResidualBlock
from src.models.stargan.discriminator import StarGANDiscriminator
from src.models.stargan.generator import StarGANGenerator

__all__ = [
    "StarGANResidualBlock",
    "StarGANGenerator",
    "StarGANDiscriminator",
]




#!/usr/bin/env python3
# Copyright (c) 2026 EarthBridge Team.
"""Convert original BBDM checkpoints to community ``unet/`` layout.

Usage:
    python -m examples.community.bbdm.convert_ckpt_to_unet \
      --raw-root "/root/worksapce/models/raw/BBDM Checkpoints" \
      --output-root "/root/worksapce/models/BiliSakura/BBDM-ckpt"
"""

from __future__ import annotations

import argparse
import json
from pathlib import Path
from typing import Any

import torch
from safetensors.torch import save_file


def _load_yaml(path: Path) -> dict[str, Any]:
    try:
        import yaml
    except ImportError as exc:
        raise RuntimeError(
            "PyYAML is required for BBDM conversion. Install with: pip install pyyaml"
        ) from exc
    with open(path, encoding="utf-8") as f:
        return yaml.load(f, Loader=yaml.FullLoader)


def _try_list_checkpoints(folder: Path) -> list[Path]:
    patterns = ("*.pt", "*.pth", "*.ckpt")
    out: list[Path] = []
    for pat in patterns:
        out.extend(sorted(folder.glob(pat)))
    return out


def _extract_model_state(raw_state: Any) -> dict[str, torch.Tensor]:
    if not isinstance(raw_state, dict):
        raise TypeError(f"Unexpected checkpoint object type: {type(raw_state)}")

    state = raw_state
    for key in ("state_dict", "model", "ema_model"):
        if key in state and isinstance(state[key], dict):
            state = state[key]

    if not isinstance(state, dict):
        raise TypeError("Could not resolve a state_dict from checkpoint")

    denoise_prefixes = (
        "denoise_fn.",
        "model.denoise_fn.",
        "module.denoise_fn.",
        "model.module.denoise_fn.",
    )
    openai_roots = ("input_blocks.", "middle_block.", "output_blocks.", "out.", "time_embed.")

    extracted: dict[str, torch.Tensor] = {}
    for key, value in state.items():
        new_key = None
        for prefix in denoise_prefixes:
            if key.startswith(prefix):
                new_key = key[len(prefix) :]
                break
        if new_key is None:
            if key.startswith(openai_roots):
                new_key = key
            elif key.startswith("module.") and key[len("module.") :].startswith(openai_roots):
                new_key = key[len("module.") :]
        if new_key is not None and torch.is_tensor(value):
            extracted[new_key] = value.detach().cpu().contiguous()

    if not extracted:
        # fallback: checkpoint may already be bare UNet state_dict
        if any(k.startswith(openai_roots) for k in state.keys()):
            extracted = {
                k: v.detach().cpu().contiguous()
                for k, v in state.items()
                if torch.is_tensor(v) and k.startswith(openai_roots)
            }
    if not extracted:
        raise RuntimeError(
            "No OpenAI-style denoise_fn weights found. Expected keys like "
            "'denoise_fn.input_blocks.0.0.weight'."
        )
    return extracted


def _attention_to_ds(attn: list[int] | tuple[int, ...], image_size: int) -> list[int]:
    out: list[int] = []
    for v in attn:
        if v <= 0:
            continue
        # Original configs often store resolutions (32,16,8). Convert to ds (2,4,8).
        if v > 8 and image_size % v == 0:
            out.append(image_size // v)
        else:
            out.append(v)
    # preserve order, remove duplicates
    seen = set()
    uniq = []
    for v in out:
        if v not in seen:
            uniq.append(v)
            seen.add(v)
    return uniq or [2, 4, 8]


def _infer_attention_ds_from_keys(
    *,
    denoise_keys: list[str],
    channel_mult: list[int],
    num_res_blocks: int,
) -> list[int]:
    """Infer UNet attention ds locations directly from OpenAI-style weight keys."""
    if not denoise_keys:
        return []

    input_block_ds: dict[int, int] = {}
    idx = 1  # input_blocks.0 is the initial conv block
    ds = 1
    for level, _ in enumerate(channel_mult):
        for _ in range(num_res_blocks):
            input_block_ds[idx] = ds
            idx += 1
        if level != len(channel_mult) - 1:
            idx += 1  # downsample block
            ds *= 2

    output_block_ds: dict[int, int] = {}
    ds = 2 ** (len(channel_mult) - 1)
    out_idx = 0
    for rev_level, _ in enumerate(channel_mult[::-1]):
        for i in range(num_res_blocks + 1):
            output_block_ds[out_idx] = ds
            if rev_level != len(channel_mult) - 1 and i == num_res_blocks:
                ds //= 2
            out_idx += 1

    attn_ds: list[int] = []
    seen: set[int] = set()
    for key in denoise_keys:
        m = None
        if ".1.qkv.weight" in key and key.startswith("input_blocks."):
            m = key.split(".", 2)[1]
            if m.isdigit():
                block_idx = int(m)
                ds_val = input_block_ds.get(block_idx)
                if ds_val is not None and ds_val not in seen:
                    seen.add(ds_val)
                    attn_ds.append(ds_val)
        elif ".1.qkv.weight" in key and key.startswith("output_blocks."):
            m = key.split(".", 2)[1]
            if m.isdigit():
                block_idx = int(m)
                ds_val = output_block_ds.get(block_idx)
                if ds_val is not None and ds_val not in seen:
                    seen.add(ds_val)
                    attn_ds.append(ds_val)
    return attn_ds


def _build_unet_config(
    yaml_cfg: dict[str, Any],
    ckpt_name: str,
    *,
    denoise_keys: list[str] | None = None,
) -> dict[str, Any]:
    bb_params = yaml_cfg["model"]["BB"]["params"]
    unet = bb_params["UNetParams"]
    image_size = int(unet.get("image_size", 64))
    attn = unet.get("attention_resolutions", [32, 16, 8])
    channel_mult = list(unet.get("channel_mult", [1, 4, 8]))
    num_res_blocks = int(unet.get("num_res_blocks", 2))
    attn_ds = _attention_to_ds(list(attn), image_size=image_size)
    if denoise_keys is not None:
        inferred = _infer_attention_ds_from_keys(
            denoise_keys=denoise_keys,
            channel_mult=channel_mult,
            num_res_blocks=num_res_blocks,
        )
        if inferred:
            attn_ds = inferred
        else:
            # Many LBBDM checkpoints only keep middle attention.
            attn_ds = []

    raw_condition_key = unet.get("condition_key", "nocond")
    if raw_condition_key is None or str(raw_condition_key).strip() == "":
        condition_key = "nocond"
    else:
        condition_key = str(raw_condition_key)
    condition_mode = "concat" if condition_key != "nocond" else "none"
    return {
        "image_size": image_size,
        "in_channels": int(unet.get("in_channels", 3)),
        "model_channels": int(unet.get("model_channels", 128)),
        "out_channels": int(unet.get("out_channels", 3)),
        "num_res_blocks": num_res_blocks,
        "attention_resolutions": attn_ds,
        "dropout": float(unet.get("dropout", 0.0)),
        "channel_mult": channel_mult,
        "conv_resample": bool(unet.get("conv_resample", True)),
        "use_scale_shift_norm": bool(unet.get("use_scale_shift_norm", True)),
        "num_heads": int(unet.get("num_heads", 8)),
        "num_head_channels": int(unet.get("num_head_channels", 64)),
        "resblock_updown": bool(unet.get("resblock_updown", True)),
        "use_spatial_transformer": bool(unet.get("use_spatial_transformer", False)),
        "condition_key": condition_key,
        "condition_mode": condition_mode,
        "_class_name": "OpenAIBBDMUNet",
        "_converted_from": ckpt_name,
    }


def _build_scheduler_config(yaml_cfg: dict[str, Any]) -> dict[str, Any]:
    p = yaml_cfg["model"]["BB"]["params"]
    return {
        "num_timesteps": int(p.get("num_timesteps", 1000)),
        "mt_type": str(p.get("mt_type", "linear")),
        "eta": float(p.get("eta", 1.0)),
        "max_var": float(p.get("max_var", 1.0)),
        "skip_sample": bool(p.get("skip_sample", True)),
        "sample_step": int(p.get("sample_step", 200)),
        "sample_step_type": str(p.get("sample_type", "linear")),
        "objective": str(p.get("objective", "grad")),
    }


def _check_src_bbdm_compat(weights: dict[str, torch.Tensor]) -> tuple[bool, str]:
    # src BBDMUNet expects diffusers-style keys under `unet.*` rather than OpenAI-style.
    openai_like = any(k.startswith("input_blocks.") for k in weights.keys())
    if openai_like:
        return (
            False,
            "OpenAI-style key layout detected (input_blocks/middle_block/output_blocks), "
            "not diffusers UNet2DModel format expected by src.BBDMUNet.",
        )
    return True, "No obvious incompatibility detected from key layout."


def _select_checkpoint(ckpts: list[Path], yaml_stem: str) -> Path:
    if not ckpts:
        raise ValueError("No checkpoints to select from")
    stem_lower = yaml_stem.lower()
    for ckpt in ckpts:
        if stem_lower in ckpt.stem.lower():
            return ckpt
    return ckpts[0]


def convert_one(raw_subdir: Path, yaml_path: Path, output_root: Path) -> None:
    yaml_stem = yaml_path.stem
    if "-f" in yaml_stem:
        name = yaml_stem
    else:
        name = raw_subdir.name
    out_dir = output_root / name
    out_unet = out_dir / "unet"
    out_sched = out_dir / "scheduler"
    out_unet.mkdir(parents=True, exist_ok=True)
    out_sched.mkdir(parents=True, exist_ok=True)

    yaml_cfg = _load_yaml(yaml_path)
    model_cfg = yaml_cfg.get("model", {})
    if "BB" not in model_cfg:
        print(f"[skip] {yaml_path} does not look like a BBDM/LBBDM config.")
        return
    ckpts = _try_list_checkpoints(raw_subdir)

    scheduler_cfg = _build_scheduler_config(yaml_cfg)
    with open(out_sched / "scheduler_config.json", "w", encoding="utf-8") as f:
        json.dump(scheduler_cfg, f, indent=2)

    if not ckpts:
        status = {
            "status": "pending_weights",
            "message": "No checkpoint file (*.pt/*.pth/*.ckpt) found in source folder.",
            "source_folder": str(raw_subdir),
            "expected_output": str(out_unet / "diffusion_pytorch_model.safetensors"),
        }
        with open(out_dir / "conversion_status.json", "w", encoding="utf-8") as f:
            json.dump(status, f, indent=2)
        # still write a config inferred from YAML so layout is ready.
        with open(out_unet / "config.json", "w", encoding="utf-8") as f:
            json.dump(_build_unet_config(yaml_cfg, ckpt_name="N/A"), f, indent=2)
        print(f"[{name}] no checkpoint found; wrote scaffold config + scheduler.")
        return

    ckpt = _select_checkpoint(ckpts, yaml_stem=yaml_stem)
    state = torch.load(str(ckpt), map_location="cpu", weights_only=False)
    denoise = _extract_model_state(state)

    unet_cfg = _build_unet_config(
        yaml_cfg,
        ckpt.name,
        denoise_keys=list(denoise.keys()),
    )
    with open(out_unet / "config.json", "w", encoding="utf-8") as f:
        json.dump(unet_cfg, f, indent=2)
    save_file(denoise, str(out_unet / "diffusion_pytorch_model.safetensors"))

    compatible, reason = _check_src_bbdm_compat(denoise)
    with open(out_dir / "conversion_status.json", "w", encoding="utf-8") as f:
        json.dump(
            {
                "status": "converted",
                "source_checkpoint": str(ckpt),
                "src_bbdm_compatible": compatible,
                "src_bbdm_check_reason": reason,
                "recommended_loader": (
                    "src.BBDMPipeline.from_pretrained"
                    if compatible
                    else "examples.community.bbdm.load_bbdm_community_pipeline"
                ),
            },
            f,
            indent=2,
        )
    print(f"[{name}] converted {ckpt.name}")
    print(f"  src.BBDM compatibility: {compatible} ({reason})")


def main() -> None:
    parser = argparse.ArgumentParser(description="Convert BBDM checkpoints to unet/ format")
    parser.add_argument("--raw-root", type=str, required=True, help="Path to raw BBDM checkpoints root")
    parser.add_argument("--output-root", type=str, required=True, help="Path to converted output root")
    args = parser.parse_args()

    raw_root = Path(args.raw_root)
    output_root = Path(args.output_root)
    output_root.mkdir(parents=True, exist_ok=True)

    yaml_files = sorted(raw_root.glob("*/*.yaml"))
    if not yaml_files:
        raise FileNotFoundError(f"No YAML config files found under {raw_root}")

    for yaml_path in yaml_files:
        convert_one(yaml_path.parent, yaml_path, output_root)


if __name__ == "__main__":
    main()



"""Tests for image quality metrics."""

import torch

from src.metrics.paired import PairedImageMetricEvaluator
from src.metrics.paired.psnr import compute_psnr
from src.metrics.paired.ssim import compute_ssim


class TestPairedMetrics:
    def test_psnr_identical(self):
        x = torch.rand(2, 3, 32, 32)
        psnr = compute_psnr(x, x)
        # Identical images should have very high (or inf) PSNR
        assert psnr > 30.0 or psnr == float("inf")

    def test_ssim_identical(self):
        x = torch.rand(2, 3, 32, 32)
        ssim = compute_ssim(x, x)
        # SSIM of identical images should be close to 1
        assert ssim > 0.99

    def test_psnr_different(self):
        x = torch.rand(2, 3, 32, 32)
        y = torch.rand(2, 3, 32, 32)
        psnr = compute_psnr(x, y)
        assert isinstance(psnr, float)
        assert psnr >= 0

    def test_ssim_different(self):
        x = torch.rand(2, 3, 32, 32)
        y = torch.rand(2, 3, 32, 32)
        ssim = compute_ssim(x, y)
        assert isinstance(ssim, float)
        assert 0 <= ssim <= 1

    def test_evaluator_psnr_ssim(self):
        evaluator = PairedImageMetricEvaluator(metrics=["psnr", "ssim"])
        x = torch.rand(2, 3, 32, 32)
        y = torch.rand(2, 3, 32, 32)
        scores = evaluator(x, y)
        assert "psnr" in scores
        assert "ssim" in scores
        assert isinstance(scores["psnr"], float)
        assert isinstance(scores["ssim"], float)

    def test_evaluator_subset(self):
        evaluator = PairedImageMetricEvaluator(metrics=["psnr", "ssim"])
        x = torch.rand(2, 3, 32, 32)
        scores = evaluator(x, x, metrics=["psnr"])
        assert scores.keys() == {"psnr"}



"""Example scripts for training and inference with image translation models.

Subpackages
-----------
``community/``
    Community-contributed pipelines, each in a dedicated subfolder with
    ``model.py``, ``pipeline.py``, ``train.py``, and ``README.md``.
``i2sb/``
    I2SB (Image-to-Image Schrödinger Bridge) task configs and training loop.
``inference/``
    Unified inference script supporting all bridge-diffusion methods.
``pix2pix/``
    Pix2Pix paired image-to-image translation training.
``stegogan/``
    StegoGAN non-bijective unpaired image translation training.
``stargan/``
    StarGAN multi-domain image translation training.
``cut/``
    CUT contrastive unpaired translation training.

All training and inference examples import components from ``src/`` — no
duplicated model, scheduler, or pipeline code.
"""



"""Shared blocks for pix2pixHD generators."""

from __future__ import annotations

import functools

import torch
import torch.nn as nn


class _ResnetBlock(nn.Module):
    """A residual block used by pix2pixHD generators."""

    def __init__(
        self,
        dim: int,
        padding_type: str = "reflect",
        norm_layer: type[nn.Module] | functools.partial = nn.InstanceNorm2d,
        activation: nn.Module | None = None,
        use_dropout: bool = False,
    ) -> None:
        super().__init__()
        if activation is None:
            activation = nn.ReLU(True)
        if isinstance(norm_layer, functools.partial):
            use_bias = norm_layer.func == nn.InstanceNorm2d
        else:
            use_bias = norm_layer == nn.InstanceNorm2d

        conv_block: list[nn.Module] = []
        p = 0
        if padding_type == "reflect":
            conv_block.append(nn.ReflectionPad2d(1))
        elif padding_type == "replicate":
            conv_block.append(nn.ReplicationPad2d(1))
        elif padding_type == "zero":
            p = 1
        else:
            raise ValueError(f"Unsupported padding type: {padding_type}")

        conv_block += [
            nn.Conv2d(dim, dim, kernel_size=3, padding=p, bias=use_bias),
            norm_layer(dim),
            activation,
        ]
        if use_dropout:
            conv_block += [nn.Dropout(0.5)]

        p = 0
        if padding_type == "reflect":
            conv_block.append(nn.ReflectionPad2d(1))
        elif padding_type == "replicate":
            conv_block.append(nn.ReplicationPad2d(1))
        elif padding_type == "zero":
            p = 1

        conv_block += [
            nn.Conv2d(dim, dim, kernel_size=3, padding=p, bias=use_bias),
            norm_layer(dim),
        ]
        self.conv_block = nn.Sequential(*conv_block)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x + self.conv_block(x)




"""UNSB (Unpaired Neural Schrödinger Bridge) training and inference."""

from examples.unsb.config import UNSBConfig
from examples.unsb.train_unsb import UNSBTrainer

__all__ = ["UNSBConfig", "UNSBTrainer"]



# Copyright (c) 2026 EarthBridge Team.
# Credits: Local Diffusion from Kim et al. "Tackling Structural Hallucination in Image Translation with Local Diffusion" ECCV 2024 Oral.

"""Local Diffusion trainer for paired conditional image-to-image translation."""

from __future__ import annotations

import logging
import os
from pathlib import Path

import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from torchvision import transforms
from tqdm import tqdm

from src.data.datasets import PairedImageDataset
from src.models.local_diffusion import create_unet
from src.schedulers.local_diffusion import LocalDiffusionScheduler

from .config import LocalDiffusionConfig

logger = logging.getLogger(__name__)


def _build_transform(resolution: int):
    """Build transform: resize, to tensor."""
    return transforms.Compose([
        transforms.Resize((resolution, resolution)),
        transforms.ToTensor(),
    ])


class LocalDiffusionTrainer:
    """Training harness for Local Diffusion.

    Uses :class:`PairedImageDataset` for paired source (conditioning)
    and target (ground truth) images.  The model learns to denoise the
    target image conditioned on the source.
    """

    def __init__(self, config: LocalDiffusionConfig) -> None:
        self.config = config
        self.device = torch.device(config.device if torch.cuda.is_available() else "cpu")

        # Model
        self.model = create_unet(
            dim=config.dim,
            channels=config.channels,
            dim_mults=config.dim_mults,
            resnet_block_groups=config.resnet_block_groups,
            attn_dim_head=config.attn_dim_head,
            attn_heads=config.attn_heads,
            full_attn=config.full_attn,
            cond_in_channels=config.cond_in_channels,
            cond_filters=config.cond_filters,
            init_type=config.init_type,
            init_gain=config.init_gain,
        ).to(self.device)

        # Scheduler
        self.scheduler = LocalDiffusionScheduler(
            num_train_timesteps=config.num_train_timesteps,
            beta_schedule=config.beta_schedule,
            objective=config.objective,
            min_snr_loss_weight=config.min_snr_loss_weight,
            min_snr_gamma=config.min_snr_gamma,
        )

        # Optimizer
        self.optimizer = torch.optim.Adam(
            self.model.parameters(),
            lr=config.lr,
            betas=(config.beta1, config.beta2),
        )

    def build_dataset(
        self,
        root_source: str | Path,
        root_target: str | Path,
    ) -> PairedImageDataset:
        """Build paired dataset with transforms."""
        transform = _build_transform(self.config.resolution)
        return PairedImageDataset(
            root_source=root_source,
            root_target=root_target,
            transform_source=transform,
            transform_target=transform,
        )

    def train_step(self, source: torch.Tensor, target: torch.Tensor) -> dict:
        """Single training step.

        Parameters
        ----------
        source : Tensor (B, C, H, W)
            Source/conditioning images in [0, 1].
        target : Tensor (B, C, H, W)
            Target/ground-truth images in [0, 1].

        Returns
        -------
        dict
            Loss values for logging.
        """
        cfg = self.config
        source = source.to(self.device) * 2 - 1  # [0,1] → [-1,1]
        target = target.to(self.device) * 2 - 1

        b = target.shape[0]
        t = torch.randint(0, cfg.num_train_timesteps, (b,), device=self.device).long()

        # Forward diffusion
        noise = torch.randn_like(target)
        x_t = self.scheduler.q_sample(target, t, noise=noise)

        # Model prediction
        model_output = self.model(x_t, source, t)

        # Select target based on objective
        if cfg.objective == "pred_noise":
            loss_target = noise
        elif cfg.objective == "pred_x0":
            loss_target = target
        elif cfg.objective == "pred_v":
            loss_target = (
                self.scheduler._extract(self.scheduler.sqrt_alphas_cumprod, t, target.shape) * noise
                - self.scheduler._extract(self.scheduler.sqrt_one_minus_alphas_cumprod, t, target.shape) * target
            )

        # Compute loss
        loss = self.scheduler.compute_loss(model_output, loss_target, t)

        self.optimizer.zero_grad()
        loss.backward()
        nn.utils.clip_grad_norm_(self.model.parameters(), cfg.max_grad_norm)
        self.optimizer.step()

        return {"loss": loss.item()}

    def train(
        self,
        root_source: str | Path,
        root_target: str | Path,
    ) -> None:
        """Run full training loop."""
        cfg = self.config
        os.makedirs(cfg.save_dir, exist_ok=True)

        dataset = self.build_dataset(root_source, root_target)
        dataloader = DataLoader(
            dataset,
            batch_size=cfg.batch_size,
            shuffle=True,
            num_workers=0,
            drop_last=True,
        )

        global_step = 0
        for epoch in range(cfg.epochs):
            self.model.train()
            pbar = tqdm(dataloader, desc=f"Epoch {epoch + 1}/{cfg.epochs}")

            for batch in pbar:
                source = batch["source"]
                target = batch["target"]
                logs = self.train_step(source, target)
                global_step += 1

                if global_step % cfg.log_every == 0:
                    pbar.set_postfix(logs)
                    logger.info(
                        "step %d | loss=%.4f",
                        global_step, logs["loss"],
                    )

            if (epoch + 1) % cfg.save_every == 0:
                self.save_checkpoint(cfg.save_dir, epoch + 1)

        logger.info("Local Diffusion training complete. Checkpoints saved to %s", cfg.save_dir)

    def save_checkpoint(self, save_dir: str, epoch: int) -> None:
        """Save model checkpoint."""
        path = Path(save_dir) / f"checkpoint-epoch-{epoch}"
        path.mkdir(parents=True, exist_ok=True)

        from safetensors.torch import save_file

        model_path = path / "model"
        model_path.mkdir(exist_ok=True)
        save_file(self.model.state_dict(), model_path / "model.safetensors")

        logger.info("Saved checkpoint to %s", path)



"""SAR2Optical community pipeline (pix2pix-style SAR-to-optical translation).

Adapted from:
https://github.com/yuuIind/SAR2Optical
"""

from examples.community.sar2optical.model import (
    SAR2OpticalDiscriminator,
    SAR2OpticalGenerator,
    init_weights,
)
from examples.community.sar2optical.pipeline import (
    SAR2OpticalPipeline,
    SAR2OpticalPipelineOutput,
    load_sar2optical_pipeline,
)
from examples.community.sar2optical.train import (
    SAR2OpticalConfig,
    SAR2OpticalTrainer,
)

__all__ = [
    "SAR2OpticalGenerator",
    "SAR2OpticalDiscriminator",
    "init_weights",
    "SAR2OpticalPipeline",
    "SAR2OpticalPipelineOutput",
    "load_sar2optical_pipeline",
    "SAR2OpticalConfig",
    "SAR2OpticalTrainer",
]



"""P&R — Improved Precision and Recall (Kynkäänniemi et al., NeurIPS 2019)."""

from __future__ import annotations

import tempfile
from pathlib import Path

import torch


def compute_precision_recall(
    real_images: torch.Tensor,
    fake_images: torch.Tensor,
    device: torch.device | None = None,
    **kwargs,
) -> tuple[float, float]:
    """Compute Improved Precision and Recall via torch-fidelity.

    Saves tensors to a temp dir and calls torch_fidelity. Returns (precision,
    recall). Higher is better for both.

    Parameters
    ----------
    real_images, fake_images :
        Tensors of shape ``(N, C, H, W)`` in [0, 1].

    Returns
    -------
    tuple[float, float] :
        (precision, recall).

    References
    ----------
    .. [1] Kynkäänniemi et al., "Improved Precision and Recall Metric for
           Assessing Generative Models", NeurIPS 2019.
    """
    try:
        import torch_fidelity
    except ImportError as e:
        raise ImportError(
            "P&R requires torch-fidelity. Install with: pip install torch-fidelity"
        ) from e

    from torchvision.utils import save_image

    dev = device if device is not None else real_images.device
    real_images = (real_images * 255).clamp(0, 255).byte()
    fake_images = (fake_images * 255).clamp(0, 255).byte()

    with tempfile.TemporaryDirectory() as tmp:
        real_dir = Path(tmp) / "real"
        fake_dir = Path(tmp) / "fake"
        real_dir.mkdir()
        fake_dir.mkdir()
        for i, img in enumerate(real_images.cpu()):
            save_image(img.float() / 255, real_dir / f"{i:06d}.png")
        for i, img in enumerate(fake_images.cpu()):
            save_image(img.float() / 255, fake_dir / f"{i:06d}.png")

        metrics = torch_fidelity.calculate_metrics(
            str(fake_dir),  # input1 = generated
            str(real_dir),  # input2 = real
            prc=True,
            prc_neighborhood=5,
            verbose=False,
        )
    return (
        metrics["precision"],
        metrics["recall"],
    )


def compute_precision(
    real_images: torch.Tensor,
    fake_images: torch.Tensor,
    device: torch.device | None = None,
    **kwargs,
) -> float:
    """Improved Precision only."""
    p, _ = compute_precision_recall(real_images, fake_images, device, **kwargs)
    return p


def compute_recall(
    real_images: torch.Tensor,
    fake_images: torch.Tensor,
    device: torch.device | None = None,
    **kwargs,
) -> float:
    """Improved Recall only."""
    _, r = compute_precision_recall(real_images, fake_images, device, **kwargs)
    return r


def compute_pr_f1(
    real_images: torch.Tensor,
    fake_images: torch.Tensor,
    device: torch.device | None = None,
    **kwargs,
) -> float:
    """F1 (harmonic mean) of Improved Precision and Recall."""
    p, r = compute_precision_recall(real_images, fake_images, device, **kwargs)
    return 2 * p * r / (p + r) if (p + r) > 0 else 0.0



# Copyright (c) 2026 EarthBridge Team.
# Credits: UNSB from Kim et al. "Unpaired Image-to-Image Translation via Neural Schrödinger Bridge" ICLR 2024.

"""Configuration for UNSB (Unpaired Neural Schrödinger Bridge) training."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Optional


@dataclass
class UNSBConfig:
    """Configuration for UNSB image-to-image translation.

    Parameters
    ----------
    input_nc : int
        Number of input (source domain) channels.
    output_nc : int
        Number of output (target domain) channels.
    ngf : int
        Base number of generator filters.
    ndf : int
        Base number of discriminator filters.
    n_blocks : int
        Number of conditional ResNet blocks in the generator.
    n_mlp : int
        Number of MLP layers in the noise mapping network.
    n_layers_D : int
        Number of discriminator layers.
    normG : str
        Generator normalisation: ``instance`` | ``batch``.
    normD : str
        Discriminator normalisation.
    lambda_GAN : float
        Weight for GAN loss.
    lambda_NCE : float
        Weight for PatchNCE contrastive loss.
    lambda_SB : float
        Weight for Schrödinger Bridge loss.
    tau : float
        Entropy parameter for SDE bridge dynamics.
    num_timesteps : int
        Number of bridge refinement steps (NFE).
    nce_idt : bool
        Whether to use identity NCE loss.
    nce_layers : str
        Comma-separated layer indices for NCE feature extraction.
    nce_T : float
        Temperature for contrastive loss.
    num_patches : int
        Number of patches to sample for NCE.
    gan_mode : str
        GAN loss: ``lsgan`` | ``vanilla`` | ``wgangp``.
    lr : float
        Learning rate for G, D, E.
    beta1 : float
        Adam beta1.
    beta2 : float
        Adam beta2.
    epochs : int
        Number of training epochs.
    resolution : int
        Training resolution (crop/resize to this size).
    device : str
        Device to train on.
    save_dir : str
        Directory to save checkpoints.
    save_every : int
        Save checkpoint every N epochs.
    log_every : int
        Log metrics every N steps.
    """

    input_nc: int = 3
    output_nc: int = 3
    ngf: int = 64
    ndf: int = 64
    n_blocks: int = 9
    n_mlp: int = 3
    n_layers_D: int = 3
    normG: str = "instance"
    normD: str = "instance"
    use_dropout: bool = False
    no_antialias: bool = False
    no_antialias_up: bool = False
    init_type: str = "normal"
    init_gain: float = 0.02

    # Feature network (PatchSampleMLP)
    netF: str = "mlp_sample"
    netF_nc: int = 256

    # Loss weights
    lambda_GAN: float = 1.0
    lambda_NCE: float = 1.0
    lambda_SB: float = 1.0
    tau: float = 0.01
    num_timesteps: int = 5
    nce_idt: bool = True
    nce_layers: str = "0,4,8,12,16"
    nce_T: float = 0.07
    num_patches: int = 256
    nce_includes_all_negatives_from_minibatch: bool = False
    gan_mode: str = "lsgan"

    # Training
    lr: float = 2e-4
    beta1: float = 0.5
    beta2: float = 0.999
    epochs: int = 200
    batch_size: int = 1
    resolution: int = 256

    # Paths
    device: str = "cuda"
    save_dir: str = "./checkpoints/unsb"
    save_every: int = 10
    log_every: int = 100
    resume_from: Optional[str] = None



# Copyright (c) 2026 EarthBridge Team.
# Credits: OpenEarthMap-SAR CUT models. Architecture from Park et al. "Contrastive Learning for Unpaired Image-to-Image Translation" ECCV 2020.

"""Training stub for OpenEarthMap-SAR.

OpenEarthMap-SAR checkpoints are pre-trained. For training new CUT models
with the same architecture, use the main CUT training in
``examples/cut/train_cut.py`` with a generator compatible with
:class:`OpenEarthMapSARGenerator` (anti-aliased down/upsampling).
"""



# Copyright (c) 2026 EarthBridge Team.
# Credits: Built on open-source libraries and papers acknowledged in README.md citations.

"""LBM pixel-space inference pipeline for image-to-image translation.

Implements the LBM (Latent Bridge Matching) flow-matching sampling loop
for single-step or few-step image-to-image translation.

Reference: https://arxiv.org/abs/2503.07535
"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, List, Optional, Union

import numpy as np
import torch
from PIL import Image
from tqdm.auto import tqdm

from diffusers import DiffusionPipeline
from diffusers.utils import BaseOutput

from src.models.unet.diffusers_wrappers import LBMUNet
from src.schedulers.lbm import LBMScheduler


@dataclass
class LBMPipelineOutput(BaseOutput):
    """Output of the LBM pipeline.

    Attributes
    ----------
    images : list[PIL.Image.Image] | np.ndarray | torch.Tensor
        Translated images.
    nfe : int
        Number of function evaluations used.
    """

    images: Any
    nfe: int = 0


class LBMPipeline(DiffusionPipeline):
    """Image-to-image translation pipeline using LBM flow-matching.

    Parameters
    ----------
    unet : torch.nn.Module
        A UNet (or DiT) model for denoising.
    scheduler : LBMScheduler
        An LBM scheduler for the bridge flow-matching process.
    """

    model_cpu_offload_seq = "unet"

    def __init__(self, unet: torch.nn.Module, scheduler: LBMScheduler) -> None:
        super().__init__()
        self.register_modules(unet=unet, scheduler=scheduler)

    @classmethod
    def from_pretrained(
        cls,
        pretrained_model_name_or_path: str | Path,
        *,
        subfolder: str = "unet",
        scheduler_subfolder: str = "scheduler",
        device: str | torch.device = "cpu",
        torch_dtype: torch.dtype | None = None,
        **kwargs,
    ) -> "LBMPipeline":
        """Load LBM pipeline from local checkpoint directories."""
        unet = LBMUNet.from_pretrained(pretrained_model_name_or_path, subfolder=subfolder)
        try:
            scheduler = LBMScheduler.from_pretrained(
                pretrained_model_name_or_path, subfolder=scheduler_subfolder
            )
        except Exception:
            scheduler = LBMScheduler()
        unet = unet.eval().to(device=device)
        if torch_dtype is not None:
            unet = unet.to(dtype=torch_dtype)
        return cls(unet=unet, scheduler=scheduler)

    @property
    def device(self) -> torch.device:
        return next(self.unet.parameters()).device

    @property
    def dtype(self) -> torch.dtype:
        return next(self.unet.parameters()).dtype

    @staticmethod
    def prepare_inputs(image, device, dtype):
        """Convert inputs to ``[-1, 1]`` BCHW tensors."""
        if isinstance(image, Image.Image):
            image = [image]
        if isinstance(image, list) and isinstance(image[0], Image.Image):
            images = []
            for img in image:
                img = img.convert("RGB")
                arr = np.array(img).astype(np.float32) / 255.0
                images.append(torch.from_numpy(arr).permute(2, 0, 1))
            image = torch.stack(images)
        if isinstance(image, np.ndarray):
            image = torch.from_numpy(image)
        if image.max() > 1.0:
            image = image / 255.0
        if image.min() >= 0:
            image = image * 2 - 1
        return image.to(device=device, dtype=dtype)

    @torch.no_grad()
    def __call__(
        self,
        source_image: Union[torch.Tensor, Image.Image, List[Image.Image]],
        num_inference_steps: int = 1,
        cfg_scale: float = 1.0,
        generator: Optional[Union[torch.Generator, List[torch.Generator]]] = None,
        output_type: str = "pil",
        return_dict: bool = True,
        callback: Optional[Callable[[int, int, torch.Tensor], None]] = None,
        callback_steps: int = 1,
    ):
        """Generate images using LBM bridge flow-matching.

        Parameters
        ----------
        source_image : Tensor | PIL.Image | list[PIL.Image]
            Source/condition image(s).
        num_inference_steps : int
            Number of Euler sampling steps (default: 1).
        cfg_scale : float
            Classifier-free guidance scale (default: 1.0, disables CFG).
        output_type : str
            ``"pil"`` | ``"np"`` | ``"pt"``.
        """
        device = self.device
        dtype = self.dtype

        x_source = self.prepare_inputs(source_image, device, dtype)
        batch_size = x_source.shape[0]

        self.scheduler.set_timesteps(num_inference_steps, device=device)
        sample = x_source.clone()

        has_condition = hasattr(self.unet, "condition_mode") and self.unet.condition_mode == "concat"
        cond = x_source if has_condition else None
        use_cfg = has_condition and abs(float(cfg_scale) - 1.0) > 1e-6
        null_condition = torch.zeros_like(cond) if use_cfg else None
        nfe_per_step = 2 if use_cfg else 1
        nfe_count = 0

        progress_bar = tqdm(
            enumerate(self.scheduler.timesteps),
            total=len(self.scheduler.timesteps),
            desc="LBM Sampling",
        )

        for i, t in progress_bar:
            t_batch = t.to(device).repeat(batch_size) if isinstance(t, torch.Tensor) else torch.full(
                (batch_size,), t, device=device, dtype=torch.long,
            )

            if use_cfg:
                model_input = torch.cat([sample, sample], dim=0)
                timestep_input = torch.cat([t_batch, t_batch], dim=0)
                cond_input = torch.cat([cond, null_condition], dim=0)
                pred_batched = self.unet(model_input, timestep_input, cond=cond_input)
                pred_cond, pred_uncond = pred_batched.chunk(2, dim=0)
                pred = pred_uncond + cfg_scale * (pred_cond - pred_uncond)
            else:
                pred = self.unet(sample, t_batch, cond=cond if cond is not None else None)
            nfe_count += nfe_per_step

            result = self.scheduler.step(pred, t, sample, return_dict=True)
            sample = result.prev_sample

            if i < len(self.scheduler.timesteps) - 1:
                next_t = self.scheduler.timesteps[i + 1]
                next_t_batch = next_t.to(device).repeat(batch_size)
                next_sigmas = self.scheduler.get_sigmas(next_t_batch, n_dim=sample.ndim, device=device, dtype=dtype)
                bridge_noise = self.scheduler.bridge_noise_sigma
                sample = sample + bridge_noise * (next_sigmas * (1.0 - next_sigmas)) ** 0.5 * torch.randn_like(sample)

            if callback is not None and i % callback_steps == 0:
                callback(i, len(self.scheduler.timesteps), sample)

        images = sample.clamp(-1, 1)
        if output_type == "pil":
            images = self._convert_to_pil(images)
        elif output_type == "np":
            images = self._convert_to_numpy(images)

        if not return_dict:
            return (images, nfe_count)
        return LBMPipelineOutput(images=images, nfe=nfe_count)

    @staticmethod
    def _convert_to_pil(images: torch.Tensor) -> List[Image.Image]:
        images = (images + 1) / 2
        images = images.clamp(0, 1).cpu().permute(0, 2, 3, 1).numpy()
        images = (images * 255).round().astype(np.uint8)
        pil_images = []
        for img in images:
            if img.shape[2] == 1:
                img = img.squeeze(2)
            pil_images.append(Image.fromarray(img))
        return pil_images

    @staticmethod
    def _convert_to_numpy(images: torch.Tensor) -> np.ndarray:
        images = (images + 1) / 2
        return images.clamp(0, 1).cpu().permute(0, 2, 3, 1).numpy()



"""UNet model components and diffusers-compatible wrappers."""

from src.models.unet.i2sb_unet import I2SBUNet
from src.models.unet.unet_2d import create_model
from src.models.unet.diffusers_wrappers import (
    BBDMUNet,
    BDBMUNet,
    BiBBDMUNet,
    CDTSDEUNet,
    DBIMUNet,
    DDBMUNet,
    DDIBUNet,
    I2SBDiffusersUNet,
    LBMUNet,
)

__all__ = [
    "I2SBUNet",
    "create_model",
    "BBDMUNet",
    "BDBMUNet",
    "BiBBDMUNet",
    "CDTSDEUNet",
    "DBIMUNet",
    "DDBMUNet",
    "DDIBUNet",
    "I2SBDiffusersUNet",
    "LBMUNet",
]



"""CMMD — Metric Maximum Mean Discrepancy (Jayasumana et al., CVPR 2024)."""

from __future__ import annotations

from PIL import Image

import torch


def _mmd_rbf(x: torch.Tensor, y: torch.Tensor, sigma: float = 1.0) -> float:
    """Unbiased MMD with RBF kernel."""
    n, m = x.shape[0], y.shape[0]
    if n < 2 or m < 2:
        return float("inf")
    x = x.double()
    y = y.double()
    gamma = 1.0 / (2 * sigma**2)
    k_xx = torch.exp(-gamma * torch.cdist(x, x).pow(2))
    k_yy = torch.exp(-gamma * torch.cdist(y, y).pow(2))
    k_xy = torch.exp(-gamma * torch.cdist(x, y).pow(2))
    k_xx = (k_xx.sum() - n) / (n * (n - 1) + 1e-8)
    k_yy = (k_yy.sum() - m) / (m * (m - 1) + 1e-8)
    k_xy = k_xy.sum() / (n * m)
    return float((k_xx + k_yy - 2 * k_xy).clamp(min=0).sqrt())


def compute_cmmd(
    real_images: torch.Tensor,
    fake_images: torch.Tensor,
    model_id: str = "openai/clip-vit-base-patch32",
    device: torch.device | None = None,
    **kwargs,
) -> float:
    """Compute CMMD (CLIP Maximum Mean Discrepancy).

    Uses CLIP image embeddings instead of Inception. Better aligns with
    human judgment and text-image models.

    Parameters
    ----------
    real_images, fake_images :
        Tensors of shape ``(N, C, H, W)`` in [0, 1].
    model_id :
        HuggingFace model id for CLIP.

    Returns
    -------
    float :
        CMMD score. Lower is better.

    References
    ----------
    .. [1] Jayasumana et al., "Rethinking FID: Towards a Better Evaluation
           Metric for Image Generation", CVPR 2024.
    """
    from transformers import CLIPModel, CLIPProcessor

    dev = device if device is not None else real_images.device
    processor = CLIPProcessor.from_pretrained(model_id)
    model = CLIPModel.from_pretrained(model_id).to(dev).eval()

    def embed_batch(imgs: torch.Tensor) -> torch.Tensor:
        arr = (imgs.permute(0, 2, 3, 1).cpu().numpy() * 255).clip(0, 255).astype("uint8")
        pil = [Image.fromarray(x) for x in arr]
        inp = processor(images=pil, return_tensors="pt", padding=True)
        inp = {k: v.to(dev) for k, v in inp.items()}
        with torch.no_grad():
            return model.get_image_features(**inp).float()

    feats_real = embed_batch(real_images)
    feats_fake = embed_batch(fake_images)
    return _mmd_rbf(feats_real, feats_fake)



#!/usr/bin/env python3
# Copyright (c) 2026 EarthBridge Team.
"""Convert legacy BBDM VQGAN checkpoints to Diffusers VQModel layout.

Usage:
    conda activate rsgen
    python -m examples.community.bbdm.convert_ckpt_to_vqmodel \
      --raw-dir "/root/worksapce/models/raw/BBDM Checkpoints/vqgan_f8" \
      --raw-dir "/root/worksapce/models/raw/BBDM Checkpoints/vqgan_f16_16384" \
      --raw-dir "/root/worksapce/models/raw/BBDM Checkpoints/vqgan-f4-8192" \
      --in-place \
      --subfolder "vqvae"
"""

from __future__ import annotations

import argparse
import json
import re
import tempfile
import zipfile
from pathlib import Path
from typing import Any

import torch
from diffusers import VQModel


def _load_yaml(path: Path) -> dict[str, Any]:
    try:
        import yaml
    except ImportError as exc:
        raise RuntimeError(
            "PyYAML is required for VQGAN conversion. Install with: pip install pyyaml"
        ) from exc
    with open(path, encoding="utf-8") as f:
        return yaml.safe_load(f)


def _find_checkpoint(folder: Path) -> Path:
    preferred = ("model.ckpt", "vq-f8.ckpt", "diffusion_model.pt")
    for name in preferred:
        p = folder / name
        if p.exists():
            return p
    all_ckpts: list[Path] = []
    for pat in ("*.ckpt", "*.pt", "*.pth"):
        all_ckpts.extend(sorted(folder.glob(pat)))
    if not all_ckpts:
        raise FileNotFoundError(f"No checkpoint file found in {folder}")
    return all_ckpts[0]


def _load_model_state(ckpt_path: Path) -> dict[str, torch.Tensor]:
    raw = torch.load(str(ckpt_path), map_location="cpu", weights_only=False)
    if isinstance(raw, dict) and "state_dict" in raw and isinstance(raw["state_dict"], dict):
        state = raw["state_dict"]
    elif isinstance(raw, dict):
        state = raw
    else:
        raise TypeError(f"Unsupported checkpoint type: {type(raw)}")

    out: dict[str, torch.Tensor] = {}
    for key, value in state.items():
        if not torch.is_tensor(value):
            continue
        if key.startswith("loss.") or key.startswith("discriminator."):
            continue
        out[key] = value.detach().cpu().contiguous()
    return out


def _parse_config_from_yaml(path: Path) -> dict[str, Any]:
    cfg = _load_yaml(path)
    params = cfg["model"]["params"]
    dd = params["ddconfig"]
    ch_mult = list(dd["ch_mult"])
    attn_resolutions = set(dd.get("attn_resolutions", []))
    resolution = int(dd["resolution"])
    down_block_types = []
    up_block_types = []
    for idx in range(len(ch_mult)):
        ds = 2**idx
        cur_res = resolution // ds
        use_attn = cur_res in attn_resolutions
        down_block_types.append("AttnDownEncoderBlock2D" if use_attn else "DownEncoderBlock2D")
    for idx in range(len(ch_mult)):
        old_up_idx = len(ch_mult) - 1 - idx
        ds = 2**old_up_idx
        cur_res = resolution // ds
        use_attn = cur_res in attn_resolutions
        up_block_types.append("AttnUpDecoderBlock2D" if use_attn else "UpDecoderBlock2D")
    return {
        "in_channels": int(dd["in_channels"]),
        "out_channels": int(dd["out_ch"]),
        "down_block_types": down_block_types,
        "up_block_types": up_block_types,
        "block_out_channels": [int(dd["ch"] * mult) for mult in ch_mult],
        "layers_per_block": int(dd["num_res_blocks"]),
        "act_fn": "silu",
        "latent_channels": int(dd["z_channels"]),
        "sample_size": resolution,
        "num_vq_embeddings": int(params["n_embed"]),
        "norm_num_groups": 32,
        "vq_embed_dim": int(params["embed_dim"]),
        "scaling_factor": 0.18215,
    }


def _parse_config_from_json(path: Path) -> dict[str, Any]:
    cfg = json.loads(path.read_text(encoding="utf-8"))
    if "down_block_types" in cfg:
        return {
            "in_channels": int(cfg["in_channels"]),
            "out_channels": int(cfg.get("out_channels", cfg.get("out_ch", 3))),
            "down_block_types": list(cfg["down_block_types"]),
            "up_block_types": list(cfg["up_block_types"]),
            "block_out_channels": [int(x) for x in cfg["block_out_channels"]],
            "layers_per_block": int(cfg["layers_per_block"]),
            "act_fn": str(cfg.get("act_fn", "silu")),
            "latent_channels": int(cfg.get("latent_channels", cfg.get("z_channels", 3))),
            "sample_size": int(cfg.get("sample_size", cfg.get("resolution", 256))),
            "num_vq_embeddings": int(cfg.get("num_vq_embeddings", cfg.get("n_embed", 8192))),
            "norm_num_groups": int(cfg.get("norm_num_groups", 32)),
            "vq_embed_dim": int(cfg.get("vq_embed_dim", cfg.get("embed_dim", 3))),
            "scaling_factor": float(cfg.get("scaling_factor", 0.18215)),
        }
    ch_mult = [int(x) for x in cfg["ch_mult"]]
    attn_indices = set(int(x) for x in cfg.get("attn_resolutions", []))
    down_block_types = [
        "AttnDownEncoderBlock2D" if i in attn_indices else "DownEncoderBlock2D"
        for i in range(len(ch_mult))
    ]
    up_block_types = []
    for i in range(len(ch_mult)):
        old_i = len(ch_mult) - 1 - i
        up_block_types.append("AttnUpDecoderBlock2D" if old_i in attn_indices else "UpDecoderBlock2D")
    return {
        "in_channels": int(cfg["in_channels"]),
        "out_channels": int(cfg.get("out_channels", cfg.get("out_ch", 3))),
        "down_block_types": down_block_types,
        "up_block_types": up_block_types,
        "block_out_channels": [int(cfg["ch"] * m) for m in ch_mult],
        "layers_per_block": int(cfg["num_res_blocks"]),
        "act_fn": "silu",
        "latent_channels": int(cfg.get("latent_channels", cfg.get("z_channels", 3))),
        "sample_size": int(cfg.get("sample_size", cfg.get("resolution", 256))),
        "num_vq_embeddings": int(cfg.get("num_vq_embeddings", cfg.get("n_embed", 8192))),
        "norm_num_groups": int(cfg.get("norm_num_groups", 32)),
        "vq_embed_dim": int(cfg.get("vq_embed_dim", cfg.get("embed_dim", 3))),
        "scaling_factor": float(cfg.get("scaling_factor", 0.18215)),
    }


def _infer_config_from_state(state: dict[str, torch.Tensor]) -> dict[str, Any]:
    down_levels = sorted(
        {int(m.group(1)) for k in state for m in [re.match(r"encoder\.down\.(\d+)\.", k)] if m}
    )
    if not down_levels:
        raise RuntimeError("Unable to infer VQ config: no encoder.down.* keys found")

    base_ch = int(state["encoder.conv_in.weight"].shape[0])
    block_out_channels = []
    for level in down_levels:
        key = f"encoder.down.{level}.block.0.conv2.weight"
        block_out_channels.append(int(state[key].shape[0]))
    ch_mult = [max(c // base_ch, 1) for c in block_out_channels]

    num_res_blocks = max(
        int(m.group(2))
        for k in state
        for m in [re.match(r"encoder\.down\.(\d+)\.block\.(\d+)\.", k)]
        if m
    ) + 1

    enc_attn_old = {
        int(m.group(1)) for k in state for m in [re.match(r"encoder\.down\.(\d+)\.attn\.", k)] if m
    }
    dec_attn_old = {
        int(m.group(1)) for k in state for m in [re.match(r"decoder\.up\.(\d+)\.attn\.", k)] if m
    }
    n_blocks = len(down_levels)

    down_block_types = [
        "AttnDownEncoderBlock2D" if i in enc_attn_old else "DownEncoderBlock2D"
        for i in range(n_blocks)
    ]
    up_block_types = []
    for new_i in range(n_blocks):
        old_i = n_blocks - 1 - new_i
        up_block_types.append("AttnUpDecoderBlock2D" if old_i in dec_attn_old else "UpDecoderBlock2D")

    quant_embed = state["quantize.embedding.weight"]
    n_embed = int(quant_embed.shape[0])
    embed_dim = int(quant_embed.shape[1])
    z_channels = int(state["quant_conv.weight"].shape[1])
    in_channels = int(state["encoder.conv_in.weight"].shape[1])
    out_channels = int(state["decoder.conv_out.weight"].shape[0])

    return {
        "in_channels": in_channels,
        "out_channels": out_channels,
        "down_block_types": down_block_types,
        "up_block_types": up_block_types,
        "block_out_channels": block_out_channels,
        "layers_per_block": num_res_blocks,
        "act_fn": "silu",
        "latent_channels": z_channels,
        "sample_size": 256,
        "num_vq_embeddings": n_embed,
        "norm_num_groups": 32,
        "vq_embed_dim": embed_dim,
        "scaling_factor": 0.18215,
    }


def _load_vq_config(folder: Path, state: dict[str, torch.Tensor]) -> tuple[dict[str, Any], str]:
    yaml_cfg = folder / "config.yaml"
    json_cfg = folder / "config.json"
    if yaml_cfg.exists():
        return _parse_config_from_yaml(yaml_cfg), str(yaml_cfg)
    if json_cfg.exists():
        return _parse_config_from_json(json_cfg), str(json_cfg)
    return _infer_config_from_state(state), "inferred_from_checkpoint"


def _map_attention_part(part: str) -> str:
    if part == "norm":
        return "group_norm"
    if part == "q":
        return "to_q"
    if part == "k":
        return "to_k"
    if part == "v":
        return "to_v"
    if part == "proj_out":
        return "to_out.0"
    raise ValueError(f"Unknown attention sub-key: {part}")


def _convert_state_dict(state: dict[str, torch.Tensor], n_blocks: int) -> dict[str, torch.Tensor]:
    out: dict[str, torch.Tensor] = {}
    for key, value in state.items():
        new_key: str | None = None

        if key in {
            "encoder.conv_in.weight",
            "encoder.conv_in.bias",
            "encoder.conv_out.weight",
            "encoder.conv_out.bias",
            "decoder.conv_in.weight",
            "decoder.conv_in.bias",
            "decoder.conv_out.weight",
            "decoder.conv_out.bias",
            "quant_conv.weight",
            "quant_conv.bias",
            "post_quant_conv.weight",
            "post_quant_conv.bias",
            "quantize.embedding.weight",
        }:
            new_key = key
        elif key.startswith("encoder.norm_out."):
            new_key = key.replace("encoder.norm_out.", "encoder.conv_norm_out.")
        elif key.startswith("decoder.norm_out."):
            new_key = key.replace("decoder.norm_out.", "decoder.conv_norm_out.")
        else:
            m = re.match(r"encoder\.down\.(\d+)\.block\.(\d+)\.(.+)$", key)
            if m:
                i, j, rest = int(m.group(1)), int(m.group(2)), m.group(3)
                rest = rest.replace("nin_shortcut", "conv_shortcut")
                new_key = f"encoder.down_blocks.{i}.resnets.{j}.{rest}"

            m = m or re.match(r"encoder\.down\.(\d+)\.downsample\.conv\.(.+)$", key)
            if m and new_key is None:
                i, rest = int(m.group(1)), m.group(2)
                new_key = f"encoder.down_blocks.{i}.downsamplers.0.conv.{rest}"

            m = m or re.match(r"encoder\.down\.(\d+)\.attn\.(\d+)\.(norm|q|k|v|proj_out)\.(.+)$", key)
            if m and new_key is None:
                i, j, part, rest = int(m.group(1)), int(m.group(2)), m.group(3), m.group(4)
                mapped = _map_attention_part(part)
                new_key = f"encoder.down_blocks.{i}.attentions.{j}.{mapped}.{rest}"

            m = m or re.match(r"encoder\.mid\.block_(\d+)\.(.+)$", key)
            if m and new_key is None:
                block_idx = int(m.group(1)) - 1
                rest = m.group(2).replace("nin_shortcut", "conv_shortcut")
                new_key = f"encoder.mid_block.resnets.{block_idx}.{rest}"

            m = m or re.match(r"encoder\.mid\.attn_1\.(norm|q|k|v|proj_out)\.(.+)$", key)
            if m and new_key is None:
                part, rest = m.group(1), m.group(2)
                mapped = _map_attention_part(part)
                new_key = f"encoder.mid_block.attentions.0.{mapped}.{rest}"

            m = m or re.match(r"decoder\.mid\.block_(\d+)\.(.+)$", key)
            if m and new_key is None:
                block_idx = int(m.group(1)) - 1
                rest = m.group(2).replace("nin_shortcut", "conv_shortcut")
                new_key = f"decoder.mid_block.resnets.{block_idx}.{rest}"

            m = m or re.match(r"decoder\.mid\.attn_1\.(norm|q|k|v|proj_out)\.(.+)$", key)
            if m and new_key is None:
                part, rest = m.group(1), m.group(2)
                mapped = _map_attention_part(part)
                new_key = f"decoder.mid_block.attentions.0.{mapped}.{rest}"

            m = m or re.match(r"decoder\.up\.(\d+)\.block\.(\d+)\.(.+)$", key)
            if m and new_key is None:
                old_i, j, rest = int(m.group(1)), int(m.group(2)), m.group(3)
                new_i = n_blocks - 1 - old_i
                rest = rest.replace("nin_shortcut", "conv_shortcut")
                new_key = f"decoder.up_blocks.{new_i}.resnets.{j}.{rest}"

            m = m or re.match(r"decoder\.up\.(\d+)\.upsample\.conv\.(.+)$", key)
            if m and new_key is None:
                old_i, rest = int(m.group(1)), m.group(2)
                new_i = n_blocks - 1 - old_i
                new_key = f"decoder.up_blocks.{new_i}.upsamplers.0.conv.{rest}"

            m = m or re.match(r"decoder\.up\.(\d+)\.attn\.(\d+)\.(norm|q|k|v|proj_out)\.(.+)$", key)
            if m and new_key is None:
                old_i, j, part, rest = int(m.group(1)), int(m.group(2)), m.group(3), m.group(4)
                new_i = n_blocks - 1 - old_i
                mapped = _map_attention_part(part)
                new_key = f"decoder.up_blocks.{new_i}.attentions.{j}.{mapped}.{rest}"

        if new_key is not None:
            # Legacy taming attention uses Conv2d(1x1); Diffusers attention uses Linear.
            if new_key.endswith((".to_q.weight", ".to_k.weight", ".to_v.weight", ".to_out.0.weight")):
                if value.ndim == 4 and value.shape[-2:] == (1, 1):
                    value = value[:, :, 0, 0]
            out[new_key] = value

    return out


def convert_one(raw_dir: Path, out_dir: Path) -> None:
    ckpt_path = _find_checkpoint(raw_dir)
    state = _load_model_state(ckpt_path)
    config, config_source = _load_vq_config(raw_dir, state)
    n_blocks = len(config["block_out_channels"])
    converted = _convert_state_dict(state, n_blocks=n_blocks)

    model = VQModel(**config)
    missing, unexpected = model.load_state_dict(converted, strict=False)
    if missing:
        raise RuntimeError(f"Missing {len(missing)} keys after conversion: {missing[:10]}")
    if unexpected:
        raise RuntimeError(f"Unexpected {len(unexpected)} keys after conversion: {unexpected[:10]}")

    out_dir.mkdir(parents=True, exist_ok=True)
    model.save_pretrained(str(out_dir), safe_serialization=True)

    status = {
        "status": "converted",
        "source_checkpoint": str(ckpt_path),
        "config_source": config_source,
        "output_dir": str(out_dir),
        "class_name": "VQModel",
    }
    with open(out_dir.parent / "vqgan_conversion_status.json", "w", encoding="utf-8") as f:
        json.dump(status, f, indent=2)
    print(f"[ok] {raw_dir.name}: {ckpt_path.name} -> {out_dir}")


def _prepare_raw_source(raw_arg: str) -> tuple[Path, str, tempfile.TemporaryDirectory[str] | None]:
    """Return usable raw folder path, source name, and optional temp directory handle."""
    src = Path(raw_arg)
    if src.is_dir():
        return src, src.name, None

    if src.is_file() and src.suffix.lower() == ".zip":
        tmp = tempfile.TemporaryDirectory(prefix="vqgan_raw_")
        with zipfile.ZipFile(src, "r") as zf:
            zf.extractall(tmp.name)
        return Path(tmp.name), src.stem, tmp

    raise FileNotFoundError(f"Raw source must be a folder or .zip file: {src}")


def main() -> None:
    parser = argparse.ArgumentParser(description="Convert legacy VQGAN checkpoints to Diffusers VQModel")
    parser.add_argument(
        "--raw-dir",
        action="append",
        required=True,
        help="Raw VQGAN folder path or .zip file path",
    )
    parser.add_argument(
        "--output-root",
        type=str,
        default=None,
        help="Root output folder; if unset with --in-place, writes inside each raw folder",
    )
    parser.add_argument(
        "--in-place",
        action="store_true",
        help="Write converted VQModel under each raw folder (default subfolder: vqvae)",
    )
    parser.add_argument("--subfolder", type=str, default="vqvae", help="Subfolder name for VQModel files")
    args = parser.parse_args()

    if not args.in_place and args.output_root is None:
        raise ValueError("Specify either --in-place or --output-root")

    output_root = Path(args.output_root) if args.output_root else None
    if output_root is not None:
        output_root.mkdir(parents=True, exist_ok=True)

    for raw_dir_str in args.raw_dir:
        raw_dir, source_name, tmp_handle = _prepare_raw_source(raw_dir_str)
        try:
            if args.in_place:
                if tmp_handle is not None:
                    raise ValueError("--in-place is not supported when --raw-dir points to a .zip file")
                out_dir = raw_dir / args.subfolder
            else:
                out_dir = output_root / source_name / args.subfolder
            convert_one(raw_dir=raw_dir, out_dir=out_dir)
        finally:
            if tmp_handle is not None:
                tmp_handle.cleanup()


if __name__ == "__main__":
    main()



# Copyright (c) 2026 EarthBridge Team.
# Credits: Local Diffusion from Kim et al. "Tackling Structural Hallucination in Image Translation with Local Diffusion" ECCV 2024 Oral.

"""Configuration for Local Diffusion training."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, Tuple


@dataclass
class LocalDiffusionConfig:
    """Configuration for Local Diffusion image-to-image translation.

    Parameters
    ----------
    dim : int
        Base channel dimension for the U-Net.
    channels : int
        Number of image channels (1 for grayscale, 3 for RGB).
    dim_mults : tuple of int
        Channel multipliers for each U-Net encoder/decoder level.
    resnet_block_groups : int
        Number of groups for GroupNorm in ResNet blocks.
    attn_dim_head : int
        Dimension per attention head.
    attn_heads : int
        Number of attention heads.
    full_attn : tuple of bool
        Per-level toggle for full vs. linear attention.
    cond_use_mid : bool
        Whether the condition encoder includes a mid-level block.
    num_train_timesteps : int
        Total number of training diffusion steps.
    beta_schedule : str
        Noise schedule: ``"linear"`` | ``"cosine"`` | ``"sigmoid"``.
    objective : str
        Prediction target: ``"pred_x0"`` | ``"pred_noise"`` | ``"pred_v"``.
    branch_out : bool
        Enable branch-and-fuse during inference.
    fusion_timestep : int
        Timestep to switch from branching to fusion.
    lr : float
        Learning rate.
    epochs : int
        Number of training epochs.
    resolution : int
        Training image resolution.
    device : str
        Device to train on.
    save_dir : str
        Directory to save checkpoints.
    save_every : int
        Save checkpoint every N epochs.
    log_every : int
        Log metrics every N steps.
    """

    # Architecture
    dim: int = 32
    channels: int = 1
    dim_mults: Tuple[int, ...] = (1, 2, 4, 8)
    resnet_block_groups: int = 8
    attn_dim_head: int = 32
    attn_heads: int = 4
    full_attn: Tuple[bool, ...] = (False, False, False, True)
    cond_in_channels: Optional[int] = None
    cond_filters: Optional[Tuple[int, ...]] = None
    init_type: str = "normal"
    init_gain: float = 0.02

    # Diffusion
    num_train_timesteps: int = 250
    beta_schedule: str = "sigmoid"
    objective: str = "pred_x0"
    min_snr_loss_weight: bool = False
    min_snr_gamma: float = 5.0

    # Inference
    branch_out: bool = True
    fusion_timestep: int = 2
    ddim_sampling_eta: float = 0.0

    # Training
    lr: float = 1e-4
    beta1: float = 0.9
    beta2: float = 0.99
    max_grad_norm: float = 1.0
    epochs: int = 200
    batch_size: int = 64
    resolution: int = 28

    # Paths
    device: str = "cuda"
    save_dir: str = "./checkpoints/local_diffusion"
    save_every: int = 10
    log_every: int = 100
    resume_from: Optional[str] = None



"""Pix2Pix example: paired image-to-image translation training."""

from examples.pix2pix.train_pix2pix import Pix2PixTrainer, TrainingConfig

__all__ = [
    "Pix2PixTrainer",
    "TrainingConfig",
]



# Copyright (c) 2026 EarthBridge Team.
# Credits: Adapted from StegoGAN (Wu et al., CVPR 2024).
# Original: https://github.com/sian-wusidi/StegoGAN

"""StegoGAN network architectures for non-bijective image-to-image translation."""

from src.models.stegogan.generators import (
    ResnetMaskV1Generator,
    ResnetMaskV3Generator,
)
from src.models.stegogan.networks import (
    NetMatchability,
    ResnetBlock,
    SoftClamp,
    mask_generate,
)

__all__ = [
    "ResnetMaskV1Generator",
    "ResnetMaskV3Generator",
    "NetMatchability",
    "ResnetBlock",
    "SoftClamp",
    "mask_generate",
]



"""U-Net generator architecture."""

from __future__ import annotations

import torch
import torch.nn as nn


class UNetSkipConnection(nn.Module):
    """Single U-Net skip-connection block."""

    def __init__(
        self,
        outer_channels: int,
        inner_channels: int,
        input_channels: int | None = None,
        sub_module: nn.Module | None = None,
        outermost: bool = False,
        innermost: bool = False,
        use_dropout: bool = False,
        norm_layer: type[nn.Module] = nn.BatchNorm2d,
    ) -> None:
        super().__init__()
        self.outermost = outermost
        if input_channels is None:
            input_channels = outer_channels

        down_conv = nn.Conv2d(
            input_channels, inner_channels, kernel_size=4, stride=2, padding=1, bias=False
        )
        down_relu = nn.LeakyReLU(0.2, inplace=True)
        down_norm = norm_layer(inner_channels)
        up_relu = nn.ReLU(inplace=True)
        up_norm = norm_layer(outer_channels)

        if outermost:
            up_conv = nn.ConvTranspose2d(
                inner_channels * 2, outer_channels, kernel_size=4, stride=2, padding=1
            )
            down = [down_conv]
            up = [up_relu, up_conv, nn.Tanh()]
        elif innermost:
            up_conv = nn.ConvTranspose2d(
                inner_channels, outer_channels, kernel_size=4, stride=2, padding=1, bias=False
            )
            down = [down_relu, down_conv]
            up = [up_relu, up_conv, up_norm]
        else:
            up_conv = nn.ConvTranspose2d(
                inner_channels * 2, outer_channels, kernel_size=4, stride=2, padding=1, bias=False
            )
            down = [down_relu, down_conv, down_norm]
            up = [up_relu, up_conv, up_norm]
            if use_dropout:
                up = up + [nn.Dropout(0.5)]

        self.model: nn.Module
        if sub_module is not None:
            self.model = nn.Sequential(*down, sub_module, *up)
        else:
            self.model = nn.Sequential(*down, *up)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.outermost:
            return self.model(x)
        return torch.cat([x, self.model(x)], dim=1)


class UNetGenerator(nn.Module):
    """U-Net based generator with skip connections."""

    def __init__(
        self,
        in_channels: int = 3,
        out_channels: int = 3,
        num_downs: int = 8,
        base_filters: int = 64,
        use_dropout: bool = True,
        norm_layer: type[nn.Module] = nn.BatchNorm2d,
    ) -> None:
        super().__init__()
        block = UNetSkipConnection(
            base_filters * 8,
            base_filters * 8,
            innermost=True,
            norm_layer=norm_layer,
        )
        for _ in range(num_downs - 5):
            block = UNetSkipConnection(
                base_filters * 8,
                base_filters * 8,
                sub_module=block,
                use_dropout=use_dropout,
                norm_layer=norm_layer,
            )
        block = UNetSkipConnection(
            base_filters * 4, base_filters * 8, sub_module=block, norm_layer=norm_layer
        )
        block = UNetSkipConnection(
            base_filters * 2, base_filters * 4, sub_module=block, norm_layer=norm_layer
        )
        block = UNetSkipConnection(
            base_filters, base_filters * 2, sub_module=block, norm_layer=norm_layer
        )
        self.model = UNetSkipConnection(
            out_channels,
            base_filters,
            input_channels=in_channels,
            sub_module=block,
            outermost=True,
            norm_layer=norm_layer,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)




# Copyright (c) 2026 EarthBridge Team.
# Credits: alexzhou907/DDBM (Zhou et al., ICLR 2024).

"""DDBM community pipeline for OpenAI-style checkpoints (BiliSakura/DDBM-ckpt).

Uses :class:`OpenAIDDBMUNet` and reuses :class:`DDBMScheduler` + sampling logic
from the core DDBM pipeline.
"""

from __future__ import annotations

import json
from pathlib import Path
from typing import Union

from examples.community.ddbm.model import OpenAIDDBMUNet
from src.pipelines.ddbm import DDBMPipeline, DDBMPipelineOutput
from src.schedulers.ddbm import DDBMScheduler


def load_ddbm_community_pipeline(
    pretrained_model_name_or_path: str | Path,
    *,
    subfolder: str = "unet",
    device: str | Union[str, "torch.device"] = None,
    torch_dtype: "torch.dtype | None" = None,
) -> DDBMPipeline:
    """Load DDBM pipeline for OpenAI-style checkpoints (unet/ format).

    Use this for BiliSakura/DDBM-ckpt and other checkpoints that use
    the improved_diffusion architecture (input_blocks / middle_block / output_blocks).

    Parameters
    ----------
    pretrained_model_name_or_path : str | Path
        Path to the model directory containing unet/config.json and
        unet/diffusion_pytorch_model.safetensors.
    subfolder : str
        Subfolder containing the UNet (default ``"unet"``).
    device : str | torch.device
        Device to load the model on.
    torch_dtype : torch.dtype | None
        Optional dtype for the model.

    Returns
    -------
    DDBMPipeline
        Pipeline ready for inference.

    Example
    -------
    >>> from examples.community.ddbm import load_ddbm_community_pipeline
    >>> pipe = load_ddbm_community_pipeline("/path/to/DDBM-ckpt/edges2handbags-vp")
    >>> out = pipe(source_image=image, num_inference_steps=40, output_type="pil")
    """
    import torch

    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"
    path = Path(pretrained_model_name_or_path)
    unet = OpenAIDDBMUNet.from_pretrained(
        path, subfolder=subfolder, device=device
    )
    if torch_dtype is not None:
        unet = unet.to(dtype=torch_dtype)

    scheduler_cfg = None
    for candidate in (
        path / "scheduler" / "scheduler_config.json",
        path / "scheduler_config.json",
    ):
        if candidate.exists():
            with open(candidate, encoding="utf-8") as f:
                scheduler_cfg = json.load(f)
            break
    scheduler = (
        DDBMScheduler(**scheduler_cfg)
        if scheduler_cfg is not None
        else DDBMScheduler()
    )

    return DDBMPipeline(unet=unet, scheduler=scheduler)



# Copyright (c) 2026 EarthBridge Team.
# Credits: DiffuseIT (Kwon & Ye, ICLR 2023) - https://github.com/cyclomon/DiffuseIT

"""DiffuseIT community pipeline for diffusion-based image translation.

Loads from BiliSakura/DiffuseIT-ckpt layout (after convert_ckpt_to_diffuseit).
"""

from __future__ import annotations

import os
import shutil
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import Any, List, Optional, Union

import numpy as np
import torch
from PIL import Image

from diffusers import DiffusionPipeline
from diffusers.utils import BaseOutput

def _ensure_diffuseit_path(diffuseit_src_path: Optional[str | Path]) -> Path:
    """Resolve DiffuseIT source path."""
    if diffuseit_src_path is not None:
        path = Path(diffuseit_src_path)
        if path.exists():
            return path.resolve()
        raise FileNotFoundError(f"DiffuseIT source not found: {path}")

    candidates = [
        Path(__file__).resolve().parents[4] / "DiffuseIT",
        Path.cwd().parent / "DiffuseIT",
        Path.cwd() / "projects" / "DiffuseIT",
        Path.cwd() / "DiffuseIT",
    ]
    for p in candidates:
        if p.exists() and (p / "optimization" / "image_editor.py").exists():
            return p.resolve()

    raise FileNotFoundError(
        "DiffuseIT source not found. Clone from https://github.com/cyclomon/DiffuseIT "
        "and set diffuseit_src_path"
    )


def _setup_id_model(ckpt_dir: Path, diffuseit_path: Path) -> None:
    """Copy ArcFace id_model from self-contained ckpt to DiffuseIT for use_ffhq."""
    id_src_dir = ckpt_dir / "id_model"
    id_dst_dir = diffuseit_path / "id_model"
    id_dst_dir.mkdir(exist_ok=True)
    dst = id_dst_dir / "model_ir_se50.pth"

    src_pt = id_src_dir / "model_ir_se50.pth"
    src_safetensors = id_src_dir / "model_ir_se50.safetensors"
    if src_pt.exists():
        if not dst.exists() or dst.resolve() != src_pt.resolve():
            try:
                dst.unlink(missing_ok=True)
                dst.symlink_to(src_pt.resolve())
            except OSError:
                shutil.copy2(src_pt, dst)
    elif src_safetensors.exists():
        from safetensors.torch import load_file
        state = load_file(str(src_safetensors))
        torch.save(state, dst)


def _setup_checkpoint_for_editor(
    ckpt_dir: Path,
    diffuseit_path: Path,
    use_ffhq: bool,
    image_size: int,
) -> None:
    """Copy BiliSakura checkpoint to DiffuseIT checkpoints/ for ImageEditor."""
    ckpt_dir = Path(ckpt_dir)
    checkpoints_dir = diffuseit_path / "checkpoints"
    checkpoints_dir.mkdir(exist_ok=True)

    if use_ffhq:
        ckpt_name = "ffhq_10m.pt"
    elif image_size == 512:
        ckpt_name = "512x512_diffusion.pt"
    else:
        ckpt_name = "256x256_diffusion_uncond.pt"

    dst = checkpoints_dir / ckpt_name
    src = ckpt_dir / "diffusion_pytorch_model.pt"
    if src.exists():
        if not dst.exists() or dst.resolve() != src.resolve():
            try:
                dst.unlink(missing_ok=True)
                dst.symlink_to(src.resolve())
            except OSError:
                shutil.copy2(src, dst)
        return

    safetensors_path = ckpt_dir / "unet" / "diffusion_pytorch_model.safetensors"
    if safetensors_path.exists():
        from safetensors.torch import load_file
        state = load_file(str(safetensors_path))
        torch.save(state, dst)
        return

    raise FileNotFoundError(
        f"No checkpoint found in {ckpt_dir}. Run convert_ckpt_to_diffuseit first."
    )


@dataclass
class DiffuseITPipelineOutput(BaseOutput):
    """Output of the DiffuseIT community pipeline."""

    images: Any
    nfe: int = 0


class DiffuseITPipeline(DiffusionPipeline):
    """Image translation pipeline using DiffuseIT (ICLR 2023)."""

    def __init__(self, editor: Any, diffuseit_path: Path) -> None:
        super().__init__()
        self._editor = editor
        self._diffuseit_path = diffuseit_path

    @property
    def device(self) -> torch.device:
        return self._editor.device

    @classmethod
    def from_pretrained(
        cls,
        pretrained_model_name_or_path: str | Path,
        *,
        diffuseit_src_path: Optional[str | Path] = None,
        use_ffhq: bool = False,
        image_size: int = 256,
        timestep_respacing: str = "100",
        skip_timesteps: int = 40,
        device: Optional[str] = None,
        **kwargs,
    ) -> "DiffuseITPipeline":
        """Load DiffuseIT pipeline from BiliSakura checkpoint.

        Expects path like ``/path/to/DiffuseIT-ckpt/imagenet256-uncond`` with
        ``unet/config.json``, ``diffusion_pytorch_model.pt`` (from convert_ckpt_to_diffuseit).
        """
        ckpt_path = Path(pretrained_model_name_or_path)
        diffuseit_path = _ensure_diffuseit_path(diffuseit_src_path)

        # If ckpt_path is BiliSakura layout (not DiffuseIT repo), setup checkpoint
        if (ckpt_path / "unet" / "config.json").exists() or (
            ckpt_path / "diffusion_pytorch_model.pt"
        ).exists():
            _setup_checkpoint_for_editor(
                ckpt_path, diffuseit_path, use_ffhq, image_size
            )
            if use_ffhq:
                _setup_id_model(ckpt_path, diffuseit_path)
        elif (ckpt_path / "optimization" / "image_editor.py").exists():
            diffuseit_path = ckpt_path.resolve()

        import sys
        diffuseit_str = str(diffuseit_path)
        if diffuseit_str not in sys.path:
            sys.path.insert(0, diffuseit_str)

        from optimization.image_editor import ImageEditor

        class Args:
            pass
        args = Args()
        args.init_image = ""
        args.target_image = None
        args.prompt = ""
        args.source = ""
        args.skip_timesteps = skip_timesteps
        args.ddim = True
        args.timestep_respacing = timestep_respacing
        args.model_output_size = image_size
        args.clip_models = ["ViT-B/32"]
        args.aug_num = 8
        args.diff_iter = 50
        args.clip_guidance_lambda = 2000.0
        args.lambda_trg = 2000.0
        args.l2_trg_lambda = 3000.0
        args.range_lambda = 200.0
        args.vit_lambda = 1.0
        args.lambda_ssim = 1000.0
        args.lambda_dir_cls = 100.0
        args.lambda_contra_ssim = 200.0
        args.id_lambda = 100.0
        args.resample_num = 10
        args.seed = None
        args.gpu_id = 0
        if device:
            dev_str = str(device)
            if dev_str.startswith("cuda:") and dev_str[5:].isdigit():
                args.gpu_id = int(dev_str.split(":")[-1])
        args.output_path = tempfile.mkdtemp()
        args.output_file = "output.png"
        args.iterations_num = 1
        args.batch_size = 1
        args.use_ffhq = use_ffhq
        args.use_prog_contrast = False
        args.use_range_restart = True
        args.use_colormatch = True
        args.use_noise_aug_all = True
        args.regularize_content = False

        orig_cwd = os.getcwd()
        try:
            os.chdir(diffuseit_path)
            editor = ImageEditor(args)
        finally:
            os.chdir(orig_cwd)

        pipeline = cls(editor=editor, diffuseit_path=diffuseit_path)
        if device:
            pipeline.to(device)
        return pipeline

    def to(self, device) -> "DiffuseITPipeline":
        dev = torch.device(device) if isinstance(device, str) else device
        self._editor.device = dev
        self._editor.model.to(dev)
        if hasattr(self._editor, "VIT_LOSS") and self._editor.VIT_LOSS is not None:
            self._editor.VIT_LOSS.to(dev)
        return self

    def __call__(
        self,
        source_image: Union[Image.Image, np.ndarray, torch.Tensor],
        *,
        prompt: Optional[str] = None,
        source: Optional[str] = None,
        target_image: Optional[Union[Image.Image, np.ndarray, torch.Tensor]] = None,
        use_colormatch: bool = True,
        use_range_restart: bool = True,
        use_noise_aug_all: bool = True,
        iterations_num: int = 1,
        output_type: str = "pil",
    ) -> DiffuseITPipelineOutput:
        """Run DiffuseIT image translation."""
        if prompt is None and target_image is None:
            raise ValueError("Provide either prompt (text-guided) or target_image (image-guided)")

        pipe = self._editor
        pipe.args.use_colormatch = use_colormatch
        pipe.args.use_range_restart = use_range_restart
        pipe.args.use_noise_aug_all = use_noise_aug_all
        pipe.args.iterations_num = iterations_num
        if prompt is not None:
            pipe.args.prompt = prompt
        if source is not None:
            pipe.args.source = source
        if target_image is not None:
            pipe.args.target_image = None

        if isinstance(source_image, torch.Tensor):
            from torchvision.transforms.functional import to_pil_image
            if source_image.dim() == 3:
                source_image = source_image.unsqueeze(0)
            source_pil = to_pil_image(source_image[0].add(1).div(2).clamp(0, 1))
        elif isinstance(source_image, np.ndarray):
            source_pil = Image.fromarray(source_image).convert("RGB")
        else:
            source_pil = source_image.convert("RGB") if hasattr(source_image, "convert") else source_image

        image_size = pipe.model_config.get("image_size", 256) or 256
        source_pil = source_pil.resize((image_size, image_size), Image.LANCZOS)

        with tempfile.TemporaryDirectory() as tmpdir:
            in_path = Path(tmpdir) / "input.png"
            out_path = Path(tmpdir) / "output"
            out_path.mkdir(exist_ok=True)
            source_pil.save(in_path)
            pipe.args.init_image = str(in_path)
            pipe.args.output_path = str(out_path)
            pipe.args.output_file = "out.png"

            if target_image is not None:
                if isinstance(target_image, torch.Tensor):
                    from torchvision.transforms.functional import to_pil_image
                    t = target_image
                    if t.dim() == 3:
                        t = t.unsqueeze(0)
                    tg_pil = to_pil_image(t[0].add(1).div(2).clamp(0, 1))
                elif isinstance(target_image, np.ndarray):
                    tg_pil = Image.fromarray(target_image).convert("RGB")
                else:
                    tg_pil = target_image.convert("RGB") if hasattr(target_image, "convert") else target_image
                tg_pil = tg_pil.resize((image_size, image_size), Image.LANCZOS)
                tg_path = Path(tmpdir) / "target.png"
                tg_pil.save(tg_path)
                pipe.args.target_image = str(tg_path)
            else:
                pipe.args.target_image = None

            orig_cwd = os.getcwd()
            try:
                os.chdir(self._diffuseit_path)
                pipe.edit_image_by_prompt()
            finally:
                os.chdir(orig_cwd)

            outs = list(out_path.glob("out_i_*_b_0.png"))
            if not outs:
                outs = list(out_path.glob("*.png"))
            if not outs:
                raise RuntimeError("DiffuseIT produced no output")
            result_pil = Image.open(sorted(outs)[-1]).convert("RGB")

        total_t = getattr(pipe.diffusion, "num_timesteps", 1000)
        num_steps = total_t - pipe.args.skip_timesteps
        if pipe.args.ddim and hasattr(pipe.args, "timestep_respacing"):
            num_steps = len(
                _timestep_respacing_to_steps(pipe.args.timestep_respacing, total_t)
            )

        images: List[Any] = [result_pil]
        if output_type == "pt":
            from torchvision.transforms.functional import to_tensor
            images = [to_tensor(img).unsqueeze(0).mul(2).sub(1) for img in images]

        return DiffuseITPipelineOutput(images=images, nfe=num_steps)


def _timestep_respacing_to_steps(respacing: str, total: int) -> list:
    parts = respacing.replace(" ", "").split(",")
    if len(parts) == 1 and parts[0].isdigit():
        n = int(parts[0])
        return list(range(0, total, max(1, total // n)))
    return list(range(total))


def load_diffuseit_community_pipeline(
    checkpoint_path: str | Path,
    *,
    diffuseit_src_path: Optional[str | Path] = None,
    use_ffhq: bool = False,
    image_size: int = 256,
    device: str = "cuda",
    **kwargs,
) -> DiffuseITPipeline:
    """Load DiffuseIT community pipeline from BiliSakura checkpoint."""
    return DiffuseITPipeline.from_pretrained(
        checkpoint_path,
        diffuseit_src_path=diffuseit_src_path,
        use_ffhq=use_ffhq,
        image_size=image_size,
        device=device,
        **kwargs,
    )



"""Tests for the inference predictor."""

import torch
from PIL import Image

from src.inference.predictor import ImageTranslator
from src.models.generators import UNetGenerator


class TestImageTranslator:
    def test_predict_returns_pil(self):
        gen = UNetGenerator(in_channels=3, out_channels=3, num_downs=5, base_filters=16)
        translator = ImageTranslator(gen, device="cpu", image_size=32)
        img = Image.new("RGB", (64, 64))
        result = translator.predict(img)
        assert isinstance(result, Image.Image)

    def test_predict_batch(self):
        gen = UNetGenerator(in_channels=3, out_channels=3, num_downs=5, base_filters=16)
        translator = ImageTranslator(gen, device="cpu", image_size=32)
        imgs = [Image.new("RGB", (64, 64)) for _ in range(3)]
        results = translator.predict_batch(imgs)
        assert len(results) == 3
        assert all(isinstance(r, Image.Image) for r in results)

    def test_predict_file(self, tmp_path):
        gen = UNetGenerator(in_channels=3, out_channels=3, num_downs=5, base_filters=16)
        translator = ImageTranslator(gen, device="cpu", image_size=32)
        input_path = tmp_path / "input.png"
        output_path = tmp_path / "output.png"
        Image.new("RGB", (64, 64)).save(input_path)
        translator.predict_file(input_path, output_path)
        assert output_path.exists()
        result = Image.open(output_path)
        assert result.mode == "RGB"
